import abc
import contextlib
from typing import Any, Optional, Union, cast

import torch

from replay.data.nn import TensorMap, TensorSchema


class SasRecModel(torch.nn.Module):
    """
    SasRec model
    """

    def __init__(
        self,
        schema: TensorSchema,
        num_blocks: int = 2,
        num_heads: int = 1,
        hidden_size: int = 50,
        max_len: int = 200,
        dropout: float = 0.2,
        ti_modification: bool = False,
        time_span: int = 256,
    ) -> None:
        """
        :param schema: Tensor schema of features.
        :param num_blocks: Number of Transformer blocks.
            Default: ``2``.
        :param num_heads: Number of Attention heads.
            Default: ``1``.
        :param hidden_size: Hidden size of transformer.
            Default: ``50``.
        :param max_len: Max length of sequence.
            Default: ``200``.
        :param dropout: Dropout rate.
            Default: ``0.2``.
        :param ti_modification: Enable time relation.
            Default: ``False``.
        :param time_span: Time span if `ti_modification` is `True`.
            Default: ``256``.
        """
        super().__init__()

        # Hyperparams
        self.num_blocks = num_blocks
        self.num_heads = num_heads
        self.hidden_size = hidden_size
        self.max_len = max_len
        self.dropout = dropout
        self.ti_modification = ti_modification
        self.time_span = time_span

        item_count = schema.item_id_features.item().cardinality
        assert item_count
        self.item_count = item_count
        self.padding_idx = item_count

        assert schema.item_id_feature_name
        self.schema = schema
        self.item_feature_name = schema.item_id_feature_name

        # Model blocks
        self.masking = SasRecMasks(
            schema=schema,
            padding_idx=self.padding_idx,
        )
        self.item_embedder: Union[TiSasRecEmbeddings, SasRecEmbeddings]
        self.sasrec_layers: torch.nn.Module

        if self.ti_modification:
            self.item_embedder = TiSasRecEmbeddings(
                schema=schema,
                embed_size=self.hidden_size,
                max_len=self.max_len,
                dropout=self.dropout,
                padding_idx=self.padding_idx,
                time_span=self.time_span,
            )
            self.sasrec_layers = TiSasRecLayers(
                hidden_size=self.hidden_size,
                num_heads=self.num_heads,
                num_blocks=self.num_blocks,
                dropout=self.dropout,
            )
        else:
            self.item_embedder = SasRecEmbeddings(
                schema=schema,
                embed_size=self.hidden_size,
                max_len=self.max_len,
                dropout=self.dropout,
                padding_idx=self.padding_idx,
            )
            self.sasrec_layers = SasRecLayers(
                hidden_size=self.hidden_size,
                num_heads=self.num_heads,
                num_blocks=self.num_blocks,
                dropout=self.dropout,
            )
        self.output_normalization = SasRecNormalizer(
            hidden_size=self.hidden_size,
        )
        self._head = EmbeddingTyingHead(self.item_embedder)
        self._init()

    def forward(
        self,
        feature_tensor: TensorMap,
        padding_mask: torch.BoolTensor,
    ) -> torch.Tensor:
        """
        :param feature_tensor: Batch of features.
        :param padding_mask: Padding mask where 0 - <PAD>, 1 otherwise.

        :returns: Calculated scores.
        """
        output_embeddings = self.forward_step(feature_tensor, padding_mask)
        all_scores = self.get_logits(output_embeddings)

        return all_scores  # [B x L x E]

    def predict(
        self,
        feature_tensor: TensorMap,
        padding_mask: torch.BoolTensor,
        candidates_to_score: Optional[torch.LongTensor] = None,
    ) -> torch.Tensor:
        """
        :param feature_tensor: Batch of features.
        :param padding_mask: Padding mask where 0 - <PAD>, 1 otherwise.
        :param candidates_to_score: Item ids to calculate scores.
            if `None` predicts for all items

        :returns: Prediction among canditates_to_score items.
        """
        # final_emb: [B x E]
        final_emb = self.get_query_embeddings(feature_tensor, padding_mask)
        candidate_scores = self.get_logits(final_emb, candidates_to_score)
        return candidate_scores

    def get_query_embeddings(
        self,
        feature_tensor: TensorMap,
        padding_mask: torch.BoolTensor,
    ):
        """
        :param feature_tensor: Batch of features.
        :param padding_mask: Padding mask where 0 - <PAD>, 1 otherwise.

        :returns: Query embeddings.
        """
        return self.forward_step(feature_tensor, padding_mask)[:, -1, :]

    def forward_step(
        self,
        feature_tensor: TensorMap,
        padding_mask: torch.BoolTensor,
    ) -> torch.Tensor:
        """
        :param feature_tensor: Batch of features.
        :param padding_mask: Padding mask where 0 - <PAD>, 1 otherwise.

        :returns: Output embeddings.
        """
        device = feature_tensor[self.item_feature_name].device
        attention_mask, padding_mask, feature_tensor = self.masking(feature_tensor, padding_mask)
        if self.ti_modification:
            seqs, ti_embeddings = self.item_embedder(feature_tensor, padding_mask)
            seqs = self.sasrec_layers(seqs, attention_mask, padding_mask, ti_embeddings, device)
        else:
            seqs = self.item_embedder(feature_tensor, padding_mask)
            seqs = self.sasrec_layers(seqs, attention_mask, padding_mask)
        output_emb = self.output_normalization(seqs)

        return output_emb

    def get_logits(self, out_embeddings: torch.Tensor, item_ids: Optional[torch.LongTensor] = None) -> torch.Tensor:
        """
        Apply head to output embeddings of `forward_step`.

        :param out_embeddings: Embeddings after `forward step`.
        :param item_ids: Item ids to calculate scores.
            Default: ``None``.

        :returns: Logits for each element in `item_ids`.
        """
        return self._head(out_embeddings, item_ids)

    def _init(self) -> None:
        for _, param in self.named_parameters():
            with contextlib.suppress(ValueError):
                torch.nn.init.xavier_normal_(param.data)


class SasRecMasks:
    """
    SasRec Masks
        1. AttentionMask
        2. PaddingMask
    """

    def __init__(self, schema: TensorSchema, padding_idx: int) -> None:
        """
        :param schema: Tensor schema of features.
        :param padding_idx: Padding indices.
        """
        assert schema.item_id_feature_name
        self.item_feature_name = schema.item_id_feature_name
        self.padding_idx = padding_idx

    def __call__(
        self,
        feature_tensor: TensorMap,
        padding_mask: torch.BoolTensor,
    ) -> tuple[torch.BoolTensor, torch.BoolTensor, TensorMap]:
        """
        :param feature_tensor: Batch of features.
        :param padding_mask: Padding mask where 0 - <PAD>, 1 otherwise.

        :returns: Attention mask, unsqueezed padding mask and output feature tensor.
        """
        input_sequence = feature_tensor[self.item_feature_name]

        attention_mask = ~torch.tril(
            torch.ones((input_sequence.shape[1], input_sequence.shape[1]), dtype=torch.bool, device=padding_mask.device)
        )

        output_feature_tensor = dict(feature_tensor)
        output_feature_tensor[self.item_feature_name] = input_sequence.masked_fill(
            mask=(~padding_mask),
            value=self.padding_idx,
        )

        padding_mask_unsqueezed = padding_mask.unsqueeze(-1)

        return (
            cast(torch.BoolTensor, attention_mask),
            cast(torch.BoolTensor, padding_mask_unsqueezed),
            output_feature_tensor,
        )


class BaseSasRecEmbeddings(abc.ABC):
    """
    Base SasRec embedding class
    """

    @abc.abstractmethod
    def get_item_weights(self, indices: torch.LongTensor) -> torch.Tensor:  # pragma: no cover
        """
        :param indices: Items indices.

        :returns: Item weights for specific items.
        """

    @abc.abstractmethod
    def get_all_item_weights(self) -> torch.Tensor:  # pragma: no cover
        """
        :returns: Item weights for all items.
        """

    @abc.abstractmethod
    def get_all_embeddings(self) -> dict[str, torch.Tensor]:
        """
        :returns: copy of all embeddings presented in a layer as a dict.
        """


class EmbeddingTyingHead(torch.nn.Module):
    """
    Head that calculate logits for all item_ids given output embeddings
    """

    def __init__(self, item_embedder: BaseSasRecEmbeddings):
        """
        :param item_embedder: SasRec embedding.
        """
        super().__init__()
        self._item_embedder = item_embedder

    def forward(
        self,
        out_embeddings: torch.Tensor,
        item_ids: Optional[torch.LongTensor] = None,
    ) -> torch.Tensor:
        """
        :param out_embeddings: Embeddings after `forward step`.
        :param item_ids: Item ids to calculate scores.
            Default: ``None``.

        :returns: Calculated logits.
        """
        if item_ids is not None:
            item_embeddings = self._item_embedder.get_item_weights(item_ids)
        else:
            item_embeddings = self._item_embedder.get_all_item_weights()

        if len(item_embeddings.shape) > 2:  # global_uniform, negative sharing=False, train only
            logits = (item_embeddings * out_embeddings.unsqueeze(-2)).sum(dim=-1)
        else:
            logits = torch.matmul(out_embeddings, item_embeddings.t())
        return logits


class SasRecEmbeddings(torch.nn.Module, BaseSasRecEmbeddings):
    """
    SasRec Embedding:
        1. ItemEmbedding : normal embedding matrix
        2. PositionalEmbedding : adding positional information

    Link: https://arxiv.org/pdf/1808.09781.pdf
    """

    def __init__(
        self,
        schema: TensorSchema,
        embed_size: int,
        padding_idx: int,
        max_len: int,
        dropout: float,
    ) -> None:
        """
        :param schema Tensor schema of features.
        :param embed_size: Embedding size.
        :param padding_idx: Padding index.
        :param max_len: Max length of sequence.
        :param dropout: Dropout rate.
            Default: ``0.1``.
        """
        super().__init__()
        item_count = schema.item_id_features.item().cardinality
        assert item_count

        self.item_emb = torch.nn.Embedding(item_count + 1, embed_size, padding_idx=padding_idx)
        self.pos_emb = SasRecPositionalEmbedding(max_len=max_len, d_model=embed_size)
        self.item_emb_dropout = torch.nn.Dropout(p=dropout)

        assert schema.item_id_feature_name
        self.item_feature_name = schema.item_id_feature_name

    def forward(self, feature_tensor: TensorMap, padding_mask: torch.BoolTensor) -> torch.Tensor:
        """
        :param feature_tensor: Batch of features.
        :param padding_mask: Padding mask where 0 - <PAD>, 1 otherwise.

        :returns: Embeddings for input features.
        """
        seqs = self.item_emb(feature_tensor[self.item_feature_name]) * (self.item_emb.embedding_dim**0.5)
        seqs += self.pos_emb(seqs)
        seqs = self.item_emb_dropout(seqs)
        seqs *= padding_mask
        return seqs

    def get_item_weights(self, indices: torch.LongTensor) -> torch.Tensor:
        """
        :param indices: Items indices.

        :returns: Item weights for specific items.
        """
        return self.item_emb(indices)

    def get_all_item_weights(self) -> torch.Tensor:
        """
        :returns: Item weights for all items.
        """
        # Last one is reserved for padding, so we remove it
        return self.item_emb.weight[:-1, :]

    def get_all_embeddings(self) -> dict[str, torch.Tensor]:
        """
        :returns: copy of all embeddings presented in this layer as a dict.
        """
        return {
            "item_embedding": self.item_emb.weight.data[:-1, :].detach().clone(),
            "positional_embedding": self.pos_emb.pe.weight.data.detach().clone(),
        }


class SasRecLayers(torch.nn.Module):
    """
    SasRec vanilla layers:
        1. SelfAttention layers
        2. FeedForward layers

    Link: https://arxiv.org/pdf/1808.09781.pdf
    """

    def __init__(
        self,
        hidden_size: int,
        num_heads: int,
        num_blocks: int,
        dropout: float,
    ) -> None:
        """
        :param hidden_size: Hidden size of transformer.
        :param num_heads: Number of Attention heads.
        :param num_blocks: Number of Transformer blocks.
        :param dropout: Dropout rate.
        """
        super().__init__()
        self.attention_layers = self._layers_stacker(
            num_blocks,
            torch.nn.MultiheadAttention,
            hidden_size,
            num_heads,
            dropout,
            batch_first=True,
        )
        self.attention_layernorms = self._layers_stacker(num_blocks, torch.nn.LayerNorm, hidden_size, eps=1e-8)
        self.forward_layers = self._layers_stacker(num_blocks, SasRecPointWiseFeedForward, hidden_size, dropout)
        self.forward_layernorms = self._layers_stacker(num_blocks, torch.nn.LayerNorm, hidden_size, eps=1e-8)

    def forward(
        self,
        seqs: torch.Tensor,
        attention_mask: torch.BoolTensor,
        padding_mask: torch.BoolTensor,
    ) -> torch.Tensor:
        """
        :param seqs: Item embeddings.
        :param attention_mask: Attention mask.
        :param padding_mask: Padding mask where 0 - <PAD>, 1 otherwise.

        :returns: Output embeddings.
        """
        length = len(self.attention_layers)
        for i in range(length):
            query = self.attention_layernorms[i](seqs)
            attent_emb, _ = self.attention_layers[i](query, seqs, seqs, attn_mask=attention_mask, need_weights=False)
            seqs = query + attent_emb

            seqs = self.forward_layernorms[i](seqs)
            seqs = self.forward_layers[i](seqs)
            seqs *= padding_mask

        return seqs

    def _layers_stacker(self, num_blocks: int, layer_class: Any, *args, **kwargs) -> torch.nn.ModuleList:
        return torch.nn.ModuleList([layer_class(*args, **kwargs) for _ in range(num_blocks)])


class SasRecNormalizer(torch.nn.Module):
    """
    SasRec normalization layers

    Link: https://arxiv.org/pdf/1808.09781.pdf
    """

    def __init__(
        self,
        hidden_size: int,
    ) -> None:
        """
        :param hidden_size: Hidden size of transformer.
        """
        super().__init__()
        self.last_layernorm = torch.nn.LayerNorm(hidden_size, eps=1e-8)

    def forward(self, seqs: torch.Tensor) -> torch.Tensor:
        """
        :param seqs: Input embeddings.

        :returns: Normalized embeddings.
        """
        output_emb = self.last_layernorm(seqs)

        return output_emb


class SasRecPointWiseFeedForward(torch.nn.Module):
    """
    Point wise feed forward layers

    Link: https://arxiv.org/pdf/1808.09781.pdf
    """

    def __init__(self, hidden_units: int, dropout: float) -> None:
        """
        :param hidden_units: Hidden size.
        :param dropout: Dropout rate.
        """
        super().__init__()

        self.conv1 = torch.nn.Conv1d(hidden_units, hidden_units, kernel_size=1)
        self.dropout1 = torch.nn.Dropout(p=dropout)
        self.relu = torch.nn.ReLU()
        self.conv2 = torch.nn.Conv1d(hidden_units, hidden_units, kernel_size=1)
        self.dropout2 = torch.nn.Dropout(p=dropout)

    def forward(self, inputs: torch.LongTensor) -> torch.LongTensor:
        """
        :param inputs: Query feature vector.

        :returns: Output tensors.
        """
        outputs = self.dropout2(self.conv2(self.dropout1(self.relu(self.conv1(inputs.transpose(-1, -2))))))
        outputs = outputs.transpose(-1, -2)
        outputs += inputs

        return outputs


class SasRecPositionalEmbedding(torch.nn.Module):
    """
    Positional embedding.
    """

    def __init__(self, max_len: int, d_model: int) -> None:
        """
        :param max_len: Max sequence length.
        :param d_model: Embedding dimension.
        """
        super().__init__()
        self.pe = torch.nn.Embedding(max_len, d_model)

    def forward(self, x: torch.Tensor) -> torch.Tensor:
        """
        :param x: Input embedding.

        :returns: Positional embedding.
        """
        batch_size = x.size(0)
        return self.pe.weight.unsqueeze(0).repeat(batch_size, 1, 1)


class TiSasRecEmbeddings(torch.nn.Module, BaseSasRecEmbeddings):
    """
    TiSasRec Embedding:
        1. ItemEmbedding : normal embedding matrix
        2. TimeRelativeEmbedding: based on TiSasRec architecture
        3. TimeRelativePositionalEmbedding: based on TiSasRec architecture

    Link: https://cseweb.ucsd.edu/~jmcauley/pdfs/wsdm20b.pdf
    """

    def __init__(
        self,
        schema: TensorSchema,
        embed_size: int,
        padding_idx: int,
        max_len: int,
        time_span: int,
        dropout: float,
    ) -> None:
        """
        :param schema: Tensor schema of features.
        :param embed_size: Embedding size.
        :param padding_idx: Padding index.
        :param max_len: Max length of sequence.
        :param time_span: Time span value.
        :param dropout: Dropout rate.
        """
        super().__init__()
        self.time_span = time_span

        item_count = schema.item_id_features.item().cardinality
        assert item_count

        self.item_emb = torch.nn.Embedding(item_count + 1, embed_size, padding_idx=padding_idx)
        self.abs_pos_k_emb = SasRecPositionalEmbedding(max_len=max_len, d_model=embed_size)
        self.abs_pos_v_emb = SasRecPositionalEmbedding(max_len=max_len, d_model=embed_size)
        self.time_matrix_k_emb = torch.nn.Embedding(time_span + 1, embed_size)
        self.time_matrix_v_emb = torch.nn.Embedding(time_span + 1, embed_size)

        self.item_emb_dropout = torch.nn.Dropout(p=dropout)
        self.abs_pos_k_emb_dropout = torch.nn.Dropout(p=dropout)
        self.abs_pos_v_emb_dropout = torch.nn.Dropout(p=dropout)
        self.time_matrix_k_dropout = torch.nn.Dropout(p=dropout)
        self.time_matrix_v_dropout = torch.nn.Dropout(p=dropout)

        assert schema.item_id_feature_name
        self.item_feature_name = schema.item_id_feature_name

        assert schema.timestamp_feature_name
        self.timestamp_feature_name = schema.timestamp_feature_name

    def forward(
        self,
        feature_tensor: TensorMap,
        padding_mask: torch.BoolTensor,
    ) -> tuple[torch.Tensor, tuple[torch.Tensor, torch.Tensor, torch.Tensor, torch.Tensor]]:
        """
        :param feature_tensor: Batch of features.
        :param padding_mask: Padding mask where 0 - <PAD>, 1 otherwise.

        :returns: Output embeddings, key and value time interval matrices, key and value positional embeddings.
        """
        time_matrix = self._time_relative_matrix(feature_tensor[self.timestamp_feature_name])

        seqs = self.item_emb(feature_tensor[self.item_feature_name]) * (self.item_emb.embedding_dim**0.5)
        abs_pos_k = self.abs_pos_k_emb(seqs)
        abs_pos_v = self.abs_pos_v_emb(seqs)
        time_matrix_k = self.time_matrix_k_emb(time_matrix)
        time_matrix_v = self.time_matrix_v_emb(time_matrix)

        seqs = self.item_emb_dropout(seqs)
        seqs *= padding_mask

        abs_pos_k = self.abs_pos_k_emb_dropout(abs_pos_k)
        abs_pos_v = self.abs_pos_v_emb_dropout(abs_pos_v)
        time_matrix_k = self.time_matrix_k_dropout(time_matrix_k)
        time_matrix_v = self.time_matrix_v_dropout(time_matrix_v)

        ti_embeddings = (time_matrix_k, time_matrix_v, abs_pos_k, abs_pos_v)

        return seqs, ti_embeddings

    def _time_relative_matrix(self, time_seq: torch.Tensor) -> torch.LongTensor:
        time_matrix = torch.abs(time_seq.unsqueeze(-1) - time_seq.unsqueeze(1))
        if time_matrix.dtype != torch.int64:
            time_matrix = torch.floor(time_matrix).long()
        time_matrix = time_matrix.masked_fill(time_matrix > self.time_span, self.time_span)
        return cast(torch.LongTensor, time_matrix)

    def get_item_weights(self, indices: torch.LongTensor) -> torch.Tensor:
        """
        :param indices: Items indices.

        :returns: Item weights for specific items.
        """
        return self.item_emb(indices)

    def get_all_item_weights(self) -> torch.Tensor:
        """
        :returns: Item weights for all items.
        """
        # Last one is reserved for padding, so we remove it
        return self.item_emb.weight[:-1, :]

    def get_all_embeddings(self) -> dict[str, torch.Tensor]:
        """
        :returns: copy of all embeddings presented in this layer as a dict.
        """
        return {
            "item_embedding": self.item_emb.weight.data[:-1, :].detach().clone(),
            "abs_pos_k_emb": self.abs_pos_k_emb.pe.weight.data.detach().clone(),
            "abs_pos_v_emb": self.abs_pos_v_emb.pe.weight.data.detach().clone(),
            "time_matrix_k_emb": self.time_matrix_k_emb.weight.data.detach().clone(),
            "time_matrix_v_emb": self.time_matrix_v_emb.weight.data.detach().clone(),
        }


class TiSasRecLayers(torch.nn.Module):
    """
    TiSasRec layers:
        1. Time-relative SelfAttention layers
        2. FeedForward layers

    Link: https://cseweb.ucsd.edu/~jmcauley/pdfs/wsdm20b.pdf
    """

    def __init__(
        self,
        hidden_size: int,
        num_heads: int,
        num_blocks: int,
        dropout: float,
    ) -> None:
        """
        :param hidden_size: Hidden size of transformer.
        :param num_heads: Number of Attention heads.
        :param num_blocks: Number of Transformer blocks.
        :param dropout: Dropout rate.
        """
        super().__init__()
        self.attention_layers = self._layers_stacker(num_blocks, TiSasRecAttention, hidden_size, num_heads, dropout)
        self.forward_layers = self._layers_stacker(num_blocks, SasRecPointWiseFeedForward, hidden_size, dropout)
        self.attention_layernorms = self._layers_stacker(num_blocks, torch.nn.LayerNorm, hidden_size, eps=1e-8)
        self.forward_layernorms = self._layers_stacker(num_blocks, torch.nn.LayerNorm, hidden_size, eps=1e-8)

    def forward(
        self,
        seqs: torch.Tensor,
        attention_mask: torch.BoolTensor,
        padding_mask: torch.BoolTensor,
        ti_embeddings: tuple[torch.Tensor, torch.Tensor, torch.Tensor, torch.Tensor],
        device: torch.device,
    ) -> torch.Tensor:
        """
        :param seqs: Item embeddings.
        :param attention_mask: Attention mask.
        :param padding_mask: Padding mask where 0 - <PAD>, 1 otherwise.
        :param ti_embeddings: Output embeddings, key and value time interval matrices,
            key and value positional embeddings.
        :param device: Selected device.

        :returns: Output embeddings.
        """
        length = len(self.attention_layers)
        for i in range(length):
            query = self.attention_layernorms[i](seqs)
            attent_emb = self.attention_layers[i](query, seqs, ~padding_mask, attention_mask, ti_embeddings, device)
            seqs = query + attent_emb
            seqs = self.forward_layernorms[i](seqs)
            seqs = self.forward_layers[i](seqs)
            seqs *= padding_mask

        return seqs

    def _layers_stacker(self, num_blocks: int, layer_class: Any, *args, **kwargs) -> torch.nn.ModuleList:
        return torch.nn.ModuleList([layer_class(*args, **kwargs) for _ in range(num_blocks)])


class TiSasRecAttention(torch.nn.Module):
    """
    Time interval aware multihead attention

    Link: https://cseweb.ucsd.edu/~jmcauley/pdfs/wsdm20b.pdf
    """

    def __init__(self, hidden_size: int, head_num: int, dropout_rate: float) -> None:
        """
        :param hidden_size: Hidden embedding size.
        :param head_num: Head numbers.
        :param dropout_rate: Dropout rate.
        """
        super().__init__()
        self.query_w = torch.nn.Linear(hidden_size, hidden_size)
        self.key_w = torch.nn.Linear(hidden_size, hidden_size)
        self.value_w = torch.nn.Linear(hidden_size, hidden_size)

        self.dropout = torch.nn.Dropout(p=dropout_rate)
        self.softmax = torch.nn.Softmax(dim=-1)

        self.hidden_size = hidden_size
        self.head_num = head_num
        self.head_size = hidden_size // head_num
        self.dropout_rate = dropout_rate

    def forward(
        self,
        queries: torch.LongTensor,
        keys: torch.LongTensor,
        time_mask: torch.LongTensor,
        attn_mask: torch.LongTensor,
        ti_embeddings: tuple[torch.LongTensor, torch.LongTensor, torch.LongTensor, torch.LongTensor],
        device: torch.device,
    ) -> torch.Tensor:
        """
        :param queries: Queries feature vector.
        :param keys: Keys feature vector.
        :param time_mask: Time interval mask.
        :param attn_mask: Attention mask.
        :param ti_embeddings: Output embeddings, key and value time interval matrices,
            key and value positional embeddings..
        :param device: Selected device.

        :returns: Attention outputs.
        """
        time_matrix_k, time_matrix_v, abs_pos_k, abs_pos_v = ti_embeddings
        query, key, value = self.query_w(queries), self.key_w(keys), self.value_w(keys)

        # head dim * batch dim for parallelization (h*N, T, C/h)
        q_ = torch.cat(torch.split(query, self.head_size, dim=2), dim=0)
        k_ = torch.cat(torch.split(key, self.head_size, dim=2), dim=0)
        v_ = torch.cat(torch.split(value, self.head_size, dim=2), dim=0)

        time_matrix_k_ = torch.cat(torch.split(time_matrix_k, self.head_size, dim=3), dim=0)
        time_matrix_v_ = torch.cat(torch.split(time_matrix_v, self.head_size, dim=3), dim=0)
        abs_pos_k_ = torch.cat(torch.split(abs_pos_k, self.head_size, dim=2), dim=0)
        abs_pos_v_ = torch.cat(torch.split(abs_pos_v, self.head_size, dim=2), dim=0)

        # batched channel wise matmul to gen attention weights
        attn_weights = q_.matmul(torch.transpose(k_, 1, 2))
        attn_weights += q_.matmul(torch.transpose(abs_pos_k_, 1, 2))
        attn_weights += time_matrix_k_.matmul(q_.unsqueeze(-1)).squeeze(-1)

        # seq length adaptive scaling
        attn_weights = attn_weights / (k_.shape[-1] ** 0.5)

        t_mask = time_mask.repeat(self.head_num, 1, 1)
        t_mask = t_mask.expand(-1, -1, attn_weights.shape[-1])
        a_mask = attn_mask.unsqueeze(0).expand(attn_weights.shape[0], -1, -1)
        paddings = torch.ones(attn_weights.shape) * (-(2**32) + 1)  # -1e23 # float('-inf')
        paddings = paddings.to(device)
        attn_weights = torch.where(t_mask, paddings, attn_weights)
        attn_weights = torch.where(a_mask, paddings, attn_weights)

        attn_weights = self.softmax(attn_weights)
        attn_weights = self.dropout(attn_weights)

        outputs = attn_weights.matmul(v_)
        outputs += attn_weights.matmul(abs_pos_v_)
        outputs += attn_weights.unsqueeze(2).matmul(time_matrix_v_).reshape(outputs.shape).squeeze(2)

        outputs = torch.cat(torch.split(outputs, query.shape[0], dim=0), dim=2)

        return cast(torch.LongTensor, outputs)



from typing import Optional, Union

import pandas as pd

from .base_metric import Metric, MetricsDataFrameLike
from .offline_metrics import OfflineMetrics


class Experiment:
    """
    The class is designed for calculating, storing and comparing metrics
    from different models in the Pandas DataFrame format.

    The main difference from the ``OfflineMetrics`` class is that
    ``OfflineMetrics`` are only responsible for calculating metrics.
    The ``Experiment`` class is responsible for storing metrics from different models,
    clear and their convenient comparison with each other.

    Calculated metrics are available with ``results`` attribute.

    Example:

    >>> recommendations
       query_id  item_id  rating
    0         1        3    0.6
    1         1        7    0.5
    2         1       10    0.4
    3         1       11    0.3
    4         1        2    0.2
    5         2        5    0.6
    6         2        8    0.5
    7         2       11    0.4
    8         2        1    0.3
    9         2        3    0.2
    10        3        4    1.0
    11        3        9    0.5
    12        3        2    0.1
    >>> groundtruth
       query_id  item_id
    0         1        5
    1         1        6
    2         1        7
    3         1        8
    4         1        9
    5         1       10
    6         2        6
    7         2        7
    8         2        4
    9         2       10
    10        2       11
    11        3        1
    12        3        2
    13        3        3
    14        3        4
    15        3        5
    >>> train
       query_id  item_id
    0         1        5
    1         1        6
    2         1        8
    3         1        9
    4         1        2
    5         2        5
    6         2        8
    7         2       11
    8         2        1
    9         2        3
    10        3        4
    11        3        9
    12        3        2
    >>> base_rec
       query_id  item_id  rating
    0        1        3    0.5
    1        1        7    0.5
    2        1        2    0.7
    3        2        5    0.6
    4        2        8    0.6
    5        2        3    0.3
    6        3        4    1.0
    7        3        9    0.5
    >>> from replay.metrics import NDCG, Surprisal, Precision, Coverage, Median, ConfidenceInterval
    >>> ex = Experiment([NDCG([2, 3]), Surprisal(3)], groundtruth, train)
    >>> ex.add_result("baseline", base_rec)
    >>> ex.add_result("model", recommendations)
    >>> ex.results
                NDCG@2    NDCG@3  Surprisal@3
    baseline  0.333333  0.25512     0.608476
    model     0.333333  0.48976     0.719587
    >>> ex.compare("baseline")
              NDCG@2   NDCG@3 Surprisal@3
    baseline       –        –           –
    model     0.0%  91.97%      18.26%
    >>> ex = Experiment([Precision(3, mode=Median()), Precision(3, mode=ConfidenceInterval(0.95))], groundtruth)
    >>> ex.add_result("baseline", base_rec)
    >>> ex.add_result("model", recommendations)
    >>> ex.results
              Precision-Median@3  Precision-ConfidenceInterval@3
    baseline            0.333333                        0.217774
    model               0.666667                        0.217774
    <BLANKLINE>
    """

    def __init__(
        self,
        metrics: list[Metric],
        ground_truth: MetricsDataFrameLike,
        train: Optional[MetricsDataFrameLike] = None,
        base_recommendations: Optional[Union[MetricsDataFrameLike, dict[str, MetricsDataFrameLike]]] = None,
        query_column: str = "query_id",
        item_column: str = "item_id",
        rating_column: str = "rating",
        category_column: str = "category_id",
    ):
        """
        :param metrics: (list of metrics): List of metrics to be calculated.
        :param ground_truth: (PySpark DataFrame or Polars DataFrame or Pandas DataFrame or dict):
            test data.
            If DataFrame then it must contains user and item columns.
            If dict then key represents user_ids, value represents list of item_ids.
        :param train: (PySpark DataFrame or Polars DataFrame or Pandas DataFrame or dict, optional):
            train data.
            If DataFrame then it must contains user and item columns.
            If dict then key represents user_ids, value represents list of item_ids.
            Default: ``None``.
        :param base_recommendations: (PySpark DataFrame or Polars DataFrame or
            Pandas DataFrame or dict or Dict[str, DataFrameLike]):
            predictions from baseline model.
            If DataFrame then it must contains user, item and score columns.
            If dict then key represents user_ids, value represents list of tuple(item_id, score).
            If ``Unexpectedness`` is not in given metrics list, then you can omit this parameter.
            Default: ``None``.
        :param query_column: (str): The name of the user column.
            Note that you do not need to specify the value of this parameter for each metric separately.
            It is enough to specify the value of this parameter here once.
        :param item_column: (str): The name of the item column.
            Note that you do not need to specify the value of this parameter for each metric separately.
            It is enough to specify the value of this parameter here once.
        :param rating_column: (str): The name of the score column.
            Note that you do not need to specify the value of this parameter for each metric separately.
            It is enough to specify the value of this parameter here once.
        :param category_column: (str): The name of the category column.
            Note that you do not need to specify the value of this parameter for each metric separately.
            It is enough to specify the value of this parameter here once.

            It is used only for calculating the ``Diversity`` metric.
            If you don't calculate this metric, you can omit this parameter.
        """
        self._offline_metrics = OfflineMetrics(
            metrics=metrics,
            query_column=query_column,
            item_column=item_column,
            rating_column=rating_column,
            category_column=category_column,
        )
        self._ground_truth = ground_truth
        self._train = train
        self._base_recommendations = base_recommendations
        self.results = pd.DataFrame()

    def add_result(
        self,
        name: str,
        recommendations: MetricsDataFrameLike,
    ) -> None:
        """
        Calculate metrics for predictions

        :param name: name of the run to store in the resulting DataFrame
        :param recommendations: (PySpark DataFrame or Polars DataFrame or Pandas DataFrame or dict):
            model predictions.
            If DataFrame then it must contains user, item and score columns.
            If dict then key represents user_ids, value represents list of tuple(item_id, score).
        """
        cur_metrics = self._offline_metrics(
            recommendations, self._ground_truth, self._train, self._base_recommendations
        )
        for metric, value in cur_metrics.items():
            self.results.at[name, metric] = value

    def compare(self, name: str) -> pd.DataFrame:
        """
        Show results as a percentage difference to record ``name``.

        :param name: name of the baseline record
        :return: results table in a percentage format
        """
        if name not in self.results.index:
            msg = f"No results for model {name}"
            raise ValueError(msg)
        columns = [column for column in self.results.columns if column[-1].isdigit()]
        data_frame = self.results[columns].copy()
        baseline = data_frame.loc[name]
        for idx in data_frame.index:
            if idx != name:
                diff = data_frame.loc[idx] / baseline - 1
                data_frame.loc[idx] = [str(round(v * 100, 2)) + "%" for v in diff]
            else:
                data_frame.loc[name] = ["–"] * len(baseline)
        return data_frame



from os.path import join
from typing import Optional

from replay.data.dataset import Dataset
from replay.utils import PYSPARK_AVAILABLE, SparkDataFrame

from .base_rec import QueryRecommender

if PYSPARK_AVAILABLE:
    from pyspark.ml.clustering import KMeans, KMeansModel
    from pyspark.ml.feature import VectorAssembler
    from pyspark.sql import functions as sf


class ClusterRec(QueryRecommender):
    """
    Generate recommendations for cold queries using k-means clusters
    """

    can_predict_cold_queries = True
    _search_space = {
        "num_clusters": {"type": "int", "args": [2, 20]},
    }
    item_rel_in_cluster: SparkDataFrame

    def __init__(self, num_clusters: int = 10):
        """
        :param num_clusters: number of clusters
        """
        self.num_clusters = num_clusters

    @property
    def _init_args(self):
        return {"num_clusters": self.num_clusters}

    def _save_model(self, path: str, additional_params: Optional[dict] = None):
        super()._save_model(path, additional_params)
        self.model.write().overwrite().save(join(path, "model"))

    def _load_model(self, path: str):
        super()._load_model(path)
        self.model = KMeansModel.load(join(path, "model"))

    def _fit(
        self,
        dataset: Dataset,
    ) -> None:
        kmeans = KMeans().setK(self.num_clusters).setFeaturesCol("features")
        query_features_vector = self._transform_features(dataset.query_features)
        self.model = kmeans.fit(query_features_vector)
        queries_clusters = (
            self.model.transform(query_features_vector)
            .select(self.query_column, "prediction")
            .withColumnRenamed("prediction", "cluster")
        )

        interactions = dataset.interactions.join(queries_clusters, on=self.query_column, how="left")
        self.item_rel_in_cluster = interactions.groupBy(["cluster", self.item_column]).agg(
            sf.count(self.item_column).alias("item_count")
        )

        max_count_per_cluster = self.item_rel_in_cluster.groupby("cluster").agg(
            sf.max("item_count").alias("max_count_in_cluster")
        )
        self.item_rel_in_cluster = self.item_rel_in_cluster.join(max_count_per_cluster, on="cluster")
        self.item_rel_in_cluster = self.item_rel_in_cluster.withColumn(
            self.rating_column, sf.col("item_count") / sf.col("max_count_in_cluster")
        ).drop("item_count", "max_count_in_cluster")
        self.item_rel_in_cluster.cache().count()

    def _clear_cache(self):
        if hasattr(self, "item_rel_in_cluster"):
            self.item_rel_in_cluster.unpersist()

    @property
    def _dataframes(self):
        return {"item_rel_in_cluster": self.item_rel_in_cluster}

    def _transform_features(self, query_features):
        feature_columns = query_features.drop(self.query_column).columns
        vec = VectorAssembler(inputCols=feature_columns, outputCol="features")
        return vec.transform(query_features).select(self.query_column, "features")

    def _make_query_clusters(self, queries, query_features):
        query_cnt_in_fv = (
            query_features.select(self.query_column).distinct().join(queries.distinct(), on=self.query_column).count()
        )

        query_cnt = queries.distinct().count()

        if query_cnt_in_fv < query_cnt:
            self.logger.info(
                "%s query(s) don't have a feature vector. The results will not be calculated for them.",
                query_cnt - query_cnt_in_fv,
            )

        query_features_vector = self._transform_features(query_features.join(queries, on=self.query_column))
        return (
            self.model.transform(query_features_vector)
            .select(self.query_column, "prediction")
            .withColumnRenamed("prediction", "cluster")
        )

    def _predict(
        self,
        dataset: Dataset,
        k: int,  # noqa: ARG002
        queries: SparkDataFrame,
        items: SparkDataFrame,
        filter_seen_items: bool = True,  # noqa: ARG002
    ) -> SparkDataFrame:
        query_clusters = self._make_query_clusters(queries, dataset.query_features)
        filtered_items = self.item_rel_in_cluster.join(items, on=self.item_column)
        pred = query_clusters.join(filtered_items, on="cluster").drop("cluster")
        return pred

    def _predict_pairs(
        self,
        pairs: SparkDataFrame,
        dataset: Dataset,
    ) -> SparkDataFrame:
        query_clusters = self._make_query_clusters(pairs.select(self.query_column).distinct(), dataset.query_features)
        pairs_with_clusters = pairs.join(query_clusters, on=self.query_column)
        filtered_items = self.item_rel_in_cluster.join(pairs.select(self.item_column).distinct(), on=self.item_column)
        pred = pairs_with_clusters.join(filtered_items, on=["cluster", self.item_column]).select(
            self.query_column, self.item_column, self.rating_column
        )
        return pred



"""
Splits data into train and test
"""

from .base_splitter import Splitter, SplitterReturnType
from .cold_user_random_splitter import ColdUserRandomSplitter
from .k_folds import KFolds
from .last_n_splitter import LastNSplitter
from .new_users_splitter import NewUsersSplitter
from .random_splitter import RandomSplitter
from .ratio_splitter import RatioSplitter
from .time_splitter import TimeSplitter
from .two_stage_splitter import TwoStageSplitter



"""
Contains classes for encoding categorical data

``LabelEncodingRule`` to encode categorical data with value between 0 and n_classes-1 for single column.
    Recommended to use together with the LabelEncoder.
``LabelEncoder`` to apply multiple LabelEncodingRule to dataframe.
"""

import abc
import json
import os
import warnings
from collections.abc import Mapping, Sequence
from pathlib import Path
from typing import Literal, Optional, Union

import polars as pl

from replay.utils import (
    PYSPARK_AVAILABLE,
    DataFrameLike,
    PandasDataFrame,
    PolarsDataFrame,
    SparkDataFrame,
)

if PYSPARK_AVAILABLE:
    from pyspark.sql import Window, functions as sf  # noqa: I001
    from pyspark.sql.types import LongType, IntegerType, ArrayType
    from replay.utils.session_handler import get_spark_session

HandleUnknownStrategies = Literal["error", "use_default_value", "drop"]


class LabelEncoderTransformWarning(Warning):
    """Label encoder transform warning."""


class LabelEncoderPartialFitWarning(Warning):
    """Label encoder partial fit warning."""


class BaseLabelEncodingRule(abc.ABC):  # pragma: no cover
    """
    Interface of the label encoding rule
    """

    @property
    @abc.abstractmethod
    def column(self) -> str:
        raise NotImplementedError()

    @abc.abstractmethod
    def get_mapping(self) -> Mapping:
        raise NotImplementedError()

    @abc.abstractmethod
    def get_inverse_mapping(self) -> Mapping:
        raise NotImplementedError()

    @abc.abstractmethod
    def fit(self, df: DataFrameLike) -> "BaseLabelEncodingRule":
        raise NotImplementedError()

    @abc.abstractmethod
    def partial_fit(self, df: DataFrameLike) -> "BaseLabelEncodingRule":
        raise NotImplementedError()

    @abc.abstractmethod
    def transform(self, df: DataFrameLike) -> DataFrameLike:
        raise NotImplementedError()

    @abc.abstractmethod
    def inverse_transform(self, df: DataFrameLike) -> DataFrameLike:
        raise NotImplementedError()

    @abc.abstractmethod
    def set_default_value(self, default_value: Optional[Union[int, str]]) -> None:
        raise NotImplementedError()

    @abc.abstractmethod
    def set_handle_unknown(self, handle_unknown: HandleUnknownStrategies) -> None:
        raise NotImplementedError()


class LabelEncodingRule(BaseLabelEncodingRule):
    """
    Implementation of the encoding rule for categorical variables of PySpark, Pandas and Polars Data Frames.
    Encodes target labels with value between 0 and n_classes-1 for the given column.
    It is recommended to use together with the LabelEncoder.
    """

    _ENCODED_COLUMN_SUFFIX: str = "_encoded"
    _HANDLE_UNKNOWN_STRATEGIES = ("error", "use_default_value", "drop")
    _TRANSFORM_PERFORMANCE_THRESHOLD_FOR_PANDAS = 100_000

    def __init__(
        self,
        column: str,
        mapping: Optional[Mapping] = None,
        handle_unknown: HandleUnknownStrategies = "error",
        default_value: Optional[Union[int, str]] = None,
    ):
        """
        :param column: Name of the column to encode.
        :param mapping: Prepared mapping for the column.
            If ``None``, then fit is necessary.
            If not ``None``, then the fit call does not change the already prepared mapping.
            Default: ``None``.
        :param handle_unknown:
            When set to ``error`` an error will be raised in case an unknown label is present during transform.
            When set to ``use_default_value``, the encoded value of unknown label will be set
            to the value given for the parameter default_value.
            When set to ``drop``, the unknown labels will be dropped.
            Default: ``error``.
        :param default_value: Default value that will fill the unknown labels after transform.
            When the parameter handle_unknown is set to ``use_default_value``,
            this parameter is required and will set the encoded value of unknown labels.
            It has to be distinct from the values used to encode any of the labels in fit.
            If ``None``, then keep null.
            If ``int`` value, then fill by that value.
            If ``str`` value, should be \"last\" only, then fill by ``n_classes`` value.
            Default: ``None``.
        """
        if handle_unknown not in self._HANDLE_UNKNOWN_STRATEGIES:
            msg = f"handle_unknown should be either 'error' or 'use_default_value', got {handle_unknown}."
            raise ValueError(msg)
        self._handle_unknown = handle_unknown
        if (
            self._handle_unknown == "use_default_value"
            and default_value is not None
            and not isinstance(default_value, int)
            and default_value != "last"
        ):
            msg = "Default value should be None, int or 'last'"
            raise ValueError(msg)

        self._default_value = default_value
        self._col = column
        self._target_col = column + self._ENCODED_COLUMN_SUFFIX
        self._mapping = mapping
        self._is_fitted = False
        if self._mapping is not None:
            self._inverse_mapping = self._make_inverse_mapping()
            self._inverse_mapping_list = self._make_inverse_mapping_list()

    @property
    def column(self) -> str:
        return self._col

    def get_mapping(self) -> Mapping:
        if self._mapping is None:
            msg = "Label encoder is not fitted"
            raise RuntimeError(msg)
        return self._mapping

    def get_inverse_mapping(self) -> Mapping:
        if self._mapping is None:
            msg = "Label encoder is not fitted"
            raise RuntimeError(msg)
        return self._inverse_mapping

    def _make_inverse_mapping(self) -> Mapping:
        return {val: key for key, val in self.get_mapping().items()}

    def _make_inverse_mapping_list(self) -> list:
        inverse_mapping_list = [0 for _ in range(len(self.get_mapping()))]
        for k, value in self.get_mapping().items():
            inverse_mapping_list[value] = k
        return inverse_mapping_list

    def _fit_spark(self, df: SparkDataFrame) -> None:
        unique_col_values = df.select(self._col).distinct()
        window_function_give_ids = Window.orderBy(self._col)

        mapping_on_spark = (
            unique_col_values.withColumn(
                self._target_col,
                sf.row_number().over(window_function_give_ids).cast(LongType()),
            )
            .withColumn(self._target_col, sf.col(self._target_col) - 1)
            .select(self._col, self._target_col)
        )

        self._mapping = mapping_on_spark.rdd.collectAsMap()

    def _fit_pandas(self, df: PandasDataFrame) -> None:
        unique_col_values = df[self._col].drop_duplicates().reset_index(drop=True)
        self._mapping = {val: key for key, val in unique_col_values.to_dict().items()}

    def _fit_polars(self, df: PolarsDataFrame) -> None:
        unique_col_values = df.select(self._col).unique()
        self._mapping = {key: val for val, key in enumerate(unique_col_values.to_series().to_list())}

    def fit(self, df: DataFrameLike) -> "LabelEncodingRule":
        """
        Fits encoder to input dataframe.

        :param df: input dataframe.
        :returns: fitted EncodingRule.
        """
        if self._mapping is not None:
            return self

        if isinstance(df, PandasDataFrame):
            self._fit_pandas(df)
        elif isinstance(df, SparkDataFrame):
            self._fit_spark(df)
        elif isinstance(df, PolarsDataFrame):
            self._fit_polars(df)
        else:
            msg = f"{self.__class__.__name__} is not implemented for {type(df)}"
            raise NotImplementedError(msg)
        self._inverse_mapping = self._make_inverse_mapping()
        self._inverse_mapping_list = self._make_inverse_mapping_list()
        if self._handle_unknown == "use_default_value" and self._default_value in self._inverse_mapping:
            msg = (
                "The used value for default_value "
                f"{self._default_value} is one of the "
                "values already used for encoding the "
                "seen labels."
            )
            raise ValueError(msg)
        self._is_fitted = True
        return self

    def _partial_fit_spark(self, df: SparkDataFrame) -> None:
        assert self._mapping is not None
        max_value = sf.lit(max(self._mapping.values()) + 1)
        already_fitted = list(self._mapping.keys())
        new_values = {x[self._col] for x in df.select(self._col).distinct().collect()} - set(already_fitted)
        new_values_list = [[x] for x in new_values]
        if len(new_values_list) == 0:
            warnings.warn(
                "partial_fit will have no effect because "
                f"there are no new values in the incoming dataset at '{self.column}' column",
                LabelEncoderPartialFitWarning,
            )
            return
        new_unique_values_df: SparkDataFrame = get_spark_session().createDataFrame(new_values_list, schema=[self._col])
        window_function_give_ids = Window.orderBy(self._col)
        new_part_of_mapping = (
            new_unique_values_df.withColumn(
                self._target_col,
                sf.row_number().over(window_function_give_ids).cast(LongType()),
            )
            .withColumn(self._target_col, sf.col(self._target_col) - 1 + max_value)
            .select(self._col, self._target_col)
            .rdd.collectAsMap()
        )
        self._mapping.update(new_part_of_mapping)
        self._inverse_mapping.update({v: k for k, v in new_part_of_mapping.items()})
        self._inverse_mapping_list.extend(new_part_of_mapping.keys())

    def _partial_fit_pandas(self, df: PandasDataFrame) -> None:
        assert self._mapping is not None

        new_unique_values = set(df[self._col].tolist()) - set(self._mapping)
        if len(new_unique_values) == 0:
            warnings.warn(
                "partial_fit will have no effect because "
                f"there are no new values in the incoming dataset at '{self.column}' column",
                LabelEncoderPartialFitWarning,
            )
            return
        last_mapping_value = max(self._mapping.values())
        new_data: dict = {value: last_mapping_value + i for i, value in enumerate(new_unique_values, start=1)}
        self._mapping.update(new_data)
        self._inverse_mapping.update({v: k for k, v in new_data.items()})
        self._inverse_mapping_list.extend(new_data.keys())

    def _partial_fit_polars(self, df: PolarsDataFrame) -> None:
        assert self._mapping is not None

        new_unique_values = set(df.select(self._col).unique().to_series().to_list()) - set(self._mapping)
        if len(new_unique_values) == 0:
            warnings.warn(
                "partial_fit will have no effect because "
                f"there are no new values in the incoming dataset at '{self.column}' column",
                LabelEncoderPartialFitWarning,
            )
            return
        new_data: dict = {value: max(self._mapping.values()) + i for i, value in enumerate(new_unique_values, start=1)}
        self._mapping.update(new_data)
        self._inverse_mapping.update({v: k for k, v in new_data.items()})
        self._inverse_mapping_list.extend(new_data.keys())

    def partial_fit(self, df: DataFrameLike) -> "LabelEncodingRule":
        """
        Fits new data to already fitted encoder.

        :param df: input dataframe.
        :returns: fitted EncodingRule.
        """
        if self._mapping is None:
            return self.fit(df)

        if isinstance(df, SparkDataFrame):
            self._partial_fit_spark(df)
        elif isinstance(df, PandasDataFrame):
            self._partial_fit_pandas(df)
        elif isinstance(df, PolarsDataFrame):
            self._partial_fit_polars(df)
        else:
            msg = f"{self.__class__.__name__} is not implemented for {type(df)}"
            raise NotImplementedError(msg)

        self._is_fitted = True
        return self

    def _transform_pandas(self, df: PandasDataFrame, default_value: Optional[int]) -> PandasDataFrame:
        is_unknown_label = False
        if df.shape[0] < self._TRANSFORM_PERFORMANCE_THRESHOLD_FOR_PANDAS:  # in order to speed up
            mapping = self.get_mapping()
            joined_df = df.copy()
            mapped_column = []
            for i in df[self._col]:
                val = mapping.get(i, -1)
                is_unknown_label |= val == -1
                mapped_column.append(val)
            joined_df[self._target_col] = mapped_column
        else:
            mapping_df = PandasDataFrame(self.get_mapping().items(), columns=[self._col, self._target_col])
            joined_df = df.merge(mapping_df, how="left", on=self._col)
            unknown_mask = joined_df[self._target_col].isna()
            joined_df.loc[unknown_mask, self._target_col] = -1
            is_unknown_label |= unknown_mask.sum() > 0

        if is_unknown_label:
            unknown_mask = joined_df[self._target_col] == -1
            if self._handle_unknown == "drop":
                joined_df.drop(joined_df[unknown_mask].index, inplace=True)
                if joined_df.empty:
                    warnings.warn(
                        f"You are trying to transform dataframe with all values are unknown for {self._col}, "
                        "with `handle_unknown_strategy=drop` leads to empty dataframe",
                        LabelEncoderTransformWarning,
                    )
                joined_df[self._target_col] = joined_df[self._target_col].astype("int")
            elif self._handle_unknown == "error":
                unknown_unique_labels = joined_df[self._col][unknown_mask].unique().tolist()
                msg = f"Found unknown labels {unknown_unique_labels} in column {self._col} during transform"
                raise ValueError(msg)
            else:
                if default_value != -1:
                    joined_df[self._target_col] = joined_df[self._target_col].astype("int")
                    joined_df[self._target_col] = joined_df[self._target_col].replace({-1: default_value})

        result_df = joined_df.drop(self._col, axis=1).rename(columns={self._target_col: self._col})
        return result_df

    def _transform_spark(self, df: SparkDataFrame, default_value: Optional[int]) -> SparkDataFrame:
        mapping_on_spark = get_spark_session().createDataFrame(
            data=list(self.get_mapping().items()), schema=[self._col, self._target_col]
        )
        transformed_df = df.join(mapping_on_spark, on=self._col, how="left").withColumn(
            "unknown_mask", sf.isnull(self._target_col)
        )
        unknown_label_count = transformed_df.select(sf.sum(sf.col("unknown_mask").cast("long"))).first()[0]
        if unknown_label_count > 0:
            if self._handle_unknown == "drop":
                transformed_df = transformed_df.filter("unknown_mask == False")
                if transformed_df.rdd.isEmpty():
                    warnings.warn(
                        f"You are trying to transform dataframe with all values are unknown for {self._col}, "
                        "with `handle_unknown_strategy=drop` leads to empty dataframe",
                        LabelEncoderTransformWarning,
                    )
            elif self._handle_unknown == "error":
                collected_list = transformed_df.filter("unknown_mask == True").select(self._col).distinct().collect()
                unique_labels = [row[self._col] for row in collected_list]
                msg = f"Found unknown labels {unique_labels} in column {self._col} during transform"
                raise ValueError(msg)
            else:
                if default_value:
                    transformed_df = transformed_df.fillna({self._target_col: default_value})

        result_df = transformed_df.drop(self._col, "unknown_mask").withColumnRenamed(self._target_col, self._col)
        return result_df

    def _transform_polars(self, df: PolarsDataFrame, default_value: Optional[int]) -> SparkDataFrame:
        mapping_on_polars = pl.from_records(
            [list(self.get_mapping().keys()), list(self.get_mapping().values())],
            schema=[self._col, self._target_col],
        )
        mapping_on_polars = mapping_on_polars.with_columns(pl.col(self._col).cast(df.get_column(self._col).dtype))
        transformed_df = df.join(mapping_on_polars, on=self._col, how="left").with_columns(
            pl.col(self._target_col).is_null().alias("unknown_mask")
        )
        unknown_df = transformed_df.filter(pl.col("unknown_mask"))
        if not unknown_df.is_empty():
            if self._handle_unknown == "drop":
                transformed_df = transformed_df.filter(pl.col("unknown_mask") == "false")
                if transformed_df.is_empty():
                    warnings.warn(
                        f"You are trying to transform dataframe with all values are unknown for {self._col}, "
                        "with `handle_unknown_strategy=drop` leads to empty dataframe",
                        LabelEncoderTransformWarning,
                    )
            elif self._handle_unknown == "error":
                unique_labels = unknown_df.select(self._col).unique().to_series().to_list()
                msg = f"Found unknown labels {unique_labels} in column {self._col} during transform"
                raise ValueError(msg)
            else:
                if default_value:
                    transformed_df = transformed_df.with_columns(pl.col(self._target_col).fill_null(default_value))

        result_df = transformed_df.drop([self._col, "unknown_mask"]).rename({self._target_col: self._col})
        return result_df

    def transform(self, df: DataFrameLike) -> DataFrameLike:
        """
        Transforms input dataframe with fitted encoder.

        :param df: input dataframe.
        :returns: transformed dataframe.
        """
        if self._mapping is None:
            msg = "Label encoder is not fitted"
            raise RuntimeError(msg)

        default_value = len(self._mapping) if self._default_value == "last" else self._default_value

        if isinstance(df, PandasDataFrame):
            transformed_df = self._transform_pandas(df, default_value)
        elif isinstance(df, SparkDataFrame):
            transformed_df = self._transform_spark(df, default_value)
        elif isinstance(df, PolarsDataFrame):
            transformed_df = self._transform_polars(df, default_value)
        else:
            msg = f"{self.__class__.__name__} is not implemented for {type(df)}"
            raise NotImplementedError(msg)
        return transformed_df

    def _inverse_transform_pandas(self, df: PandasDataFrame) -> PandasDataFrame:
        dff = df.copy()
        dff[self._col] = [self._inverse_mapping_list[i] for i in df[self._col]]
        return dff

    def _inverse_transform_spark(self, df: SparkDataFrame) -> SparkDataFrame:
        mapping_on_spark = get_spark_session().createDataFrame(
            data=list(self.get_mapping().items()), schema=[self._col, self._target_col]
        )
        transformed_df = (
            df.withColumnRenamed(self._col, self._target_col)
            .join(mapping_on_spark, on=self._target_col, how="left")
            .drop(self._target_col)
        )
        return transformed_df

    def _inverse_transform_polars(self, df: PolarsDataFrame) -> PolarsDataFrame:
        mapping_on_polars = pl.from_records(
            [list(self.get_mapping().keys()), list(self.get_mapping().values())],
            schema=[self._col, self._target_col],
        )
        transformed_df = (
            df.rename({self._col: self._target_col})
            .join(mapping_on_polars, on=self._target_col, how="left")
            .drop(self._target_col)
        )
        return transformed_df

    def inverse_transform(self, df: DataFrameLike) -> DataFrameLike:
        """
        Reverse transform of transformed dataframe.

        :param df: transformed dataframe.
        :returns: initial dataframe.
        """
        if self._mapping is None:
            msg = "Label encoder is not fitted"
            raise RuntimeError(msg)

        if isinstance(df, PandasDataFrame):
            transformed_df = self._inverse_transform_pandas(df)
        elif isinstance(df, SparkDataFrame):
            transformed_df = self._inverse_transform_spark(df)
        elif isinstance(df, PolarsDataFrame):
            transformed_df = self._inverse_transform_polars(df)
        else:
            msg = f"{self.__class__.__name__} is not implemented for {type(df)}"
            raise NotImplementedError(msg)
        return transformed_df

    def set_default_value(self, default_value: Optional[Union[int, str]]) -> None:
        """
        Sets default value to deal with unknown labels.
        Used when handle_unknown_strategy is 'use_default_value'.

        :param default_value: default value.
        """
        if default_value is not None and not isinstance(default_value, int) and default_value != "last":
            msg = "Default value should be None, int or 'last'"
            raise ValueError(msg)
        self._default_value = default_value

    def set_handle_unknown(self, handle_unknown: HandleUnknownStrategies) -> None:
        """
        Sets strategy to handle unknown labels.

        :param handle_unknown: handle unknown strategy.
        """
        if handle_unknown not in self._HANDLE_UNKNOWN_STRATEGIES:
            msg = f"handle_unknown should be either 'error' or 'use_default_value', got {handle_unknown}."
            raise ValueError(msg)
        self._handle_unknown = handle_unknown

    def save(
        self,
        path: str,
    ) -> None:
        encoder_rule_dict = {}
        encoder_rule_dict["_class_name"] = self.__class__.__name__
        encoder_rule_dict["init_args"] = {
            "column": self._col,
            "mapping": self._mapping,
            "handle_unknown": self._handle_unknown,
            "default_value": self._default_value,
        }

        column_type = str(type(next(iter(self._mapping))))

        if not isinstance(column_type, (str, int, float)):  # pragma: no cover
            msg = f"LabelEncodingRule.save() is not implemented for column type {column_type}. \
Convert type to string, integer, or float."
            raise NotImplementedError(msg)

        encoder_rule_dict["fitted_args"] = {
            "target_col": self._target_col,
            "is_fitted": self._is_fitted,
            "column_type": column_type,
        }

        base_path = Path(path).with_suffix(".replay").resolve()
        if os.path.exists(base_path):  # pragma: no cover
            msg = "There is already LabelEncodingRule object saved at the given path. File will be overwrited."
            warnings.warn(msg)
        else:  # pragma: no cover
            base_path.mkdir(parents=True, exist_ok=True)

        with open(base_path / "init_args.json", "w+") as file:
            json.dump(encoder_rule_dict, file)

    @classmethod
    def load(cls, path: str) -> "LabelEncodingRule":
        base_path = Path(path).with_suffix(".replay").resolve()
        with open(base_path / "init_args.json") as file:
            encoder_rule_dict = json.loads(file.read())

        string_column_type = encoder_rule_dict["fitted_args"]["column_type"]
        if "str" in string_column_type:
            column_type = str
        elif "int" in string_column_type:
            column_type = int
        elif "float" in string_column_type:
            column_type = float

        encoder_rule_dict["init_args"]["mapping"] = {
            column_type(key): int(value) for key, value in encoder_rule_dict["init_args"]["mapping"].items()
        }

        encoding_rule = cls(**encoder_rule_dict["init_args"])
        encoding_rule._target_col = encoder_rule_dict["fitted_args"]["target_col"]
        encoding_rule._is_fitted = encoder_rule_dict["fitted_args"]["is_fitted"]
        return encoding_rule


class SequenceEncodingRule(LabelEncodingRule):
    """
    Implementation of the encoding rule for grouped categorical variables of PySpark, Pandas and Polars Data Frames.
    Grouped means that one cell of the table contains a list with categorical values.
    Encodes target labels with value between 0 and n_classes-1 for the given column.
    It is recommended to use together with the LabelEncoder.
    """

    _FAKE_INDEX_COLUMN_NAME: str = "__index__"

    def fit(self, df: DataFrameLike) -> "SequenceEncodingRule":
        """
        Fits encoder to input dataframe.

        :param df: input dataframe.
        :returns: fitted EncodingRule.
        """
        if self._mapping is not None:
            return self

        if isinstance(df, PandasDataFrame):
            self._fit_pandas(df[[self.column]].explode(self.column))
        elif isinstance(df, SparkDataFrame):
            self._fit_spark(df.select(self.column).withColumn(self.column, sf.explode(self.column)))
        elif isinstance(df, PolarsDataFrame):
            self._fit_polars(df.select(self.column).explode(self.column))
        else:
            msg = f"{self.__class__.__name__} is not implemented for {type(df)}"
            raise NotImplementedError(msg)
        self._inverse_mapping = self._make_inverse_mapping()
        self._inverse_mapping_list = self._make_inverse_mapping_list()
        if self._handle_unknown == "use_default_value" and self._default_value in self._inverse_mapping:
            msg = (
                "The used value for default_value "
                f"{self._default_value} is one of the "
                "values already used for encoding the "
                "seen labels."
            )
            raise ValueError(msg)
        self._is_fitted = True
        return self

    def partial_fit(self, df: DataFrameLike) -> "SequenceEncodingRule":
        """
        Fits new data to already fitted encoder.

        :param df: input dataframe.
        :returns: fitted EncodingRule.
        """
        if self._mapping is None:
            return self.fit(df)
        if isinstance(df, SparkDataFrame):
            self._partial_fit_spark(df.select(self.column).withColumn(self.column, sf.explode(self.column)))
        elif isinstance(df, PandasDataFrame):
            self._partial_fit_pandas(df[[self.column]].explode(self.column))
        elif isinstance(df, PolarsDataFrame):
            self._partial_fit_polars(df.select(self.column).explode(self.column))
        else:
            msg = f"{self.__class__.__name__} is not implemented for {type(df)}"
            raise NotImplementedError(msg)

        self._is_fitted = True
        return self

    def _transform_spark(self, df: SparkDataFrame, default_value: Optional[int]) -> SparkDataFrame:
        def mapper_udf(x):
            return [mapping.get(value) for value in x]  # pragma: no cover

        mapping = self.get_mapping()
        call_mapper_udf = sf.udf(mapper_udf, ArrayType(IntegerType()))
        encoded_df = df.withColumn(self._target_col, call_mapper_udf(sf.col(self.column)))

        if self._handle_unknown == "drop":
            encoded_df = encoded_df.withColumn(self._target_col, sf.filter(self._target_col, lambda x: x.isNotNull()))
            if encoded_df.select(sf.max(sf.size(self._target_col))).first()[0] == 0:
                warnings.warn(
                    f"You are trying to transform dataframe with all values are unknown for {self._col}, "
                    "with `handle_unknown_strategy=drop` leads to empty dataframe",
                    LabelEncoderTransformWarning,
                )
        elif self._handle_unknown == "error":
            if (
                encoded_df.select(sf.sum(sf.array_contains(self._target_col, -1).isNull().cast("integer"))).first()[0]
                != 0
            ):
                msg = f"Found unknown labels in column {self._col} during transform"
                raise ValueError(msg)
        else:
            if default_value:
                encoded_df = encoded_df.withColumn(
                    self._target_col,
                    sf.transform(self._target_col, lambda x: sf.when(x.isNull(), default_value).otherwise(x)),
                )

        result_df = encoded_df.drop(self._col).withColumnRenamed(self._target_col, self._col)
        return result_df

    def _transform_pandas(self, df: PandasDataFrame, default_value: Optional[int]) -> PandasDataFrame:
        mapping = self.get_mapping()
        joined_df = df.copy()
        if self._handle_unknown == "drop":
            max_array_len = 0

            def encode_func(array_col):
                nonlocal mapping, max_array_len
                res = []
                for x in array_col:
                    cur_len = 0
                    mapped = mapping.get(x)
                    if mapped is not None:
                        res.append(mapped)
                        cur_len += 1
                    max_array_len = max(max_array_len, cur_len)
                return res

            joined_df[self._target_col] = joined_df[self._col].apply(encode_func)
            if max_array_len == 0:
                warnings.warn(
                    f"You are trying to transform dataframe with all values are unknown for {self._col}, "
                    "with `handle_unknown_strategy=drop` leads to empty dataframe",
                    LabelEncoderTransformWarning,
                )
        elif self._handle_unknown == "error":
            none_count = 0

            def encode_func(array_col):
                nonlocal mapping, none_count
                res = []
                for x in array_col:
                    mapped = mapping.get(x)
                    if mapped is None:
                        none_count += 1
                    else:
                        res.append(mapped)
                return res

            joined_df[self._target_col] = joined_df[self._col].apply(encode_func)
            if none_count != 0:
                msg = f"Found unknown labels in column {self._col} during transform"
                raise ValueError(msg)
        else:

            def encode_func(array_col):
                nonlocal mapping
                return [mapping.get(x, default_value) for x in array_col]

            joined_df[self._target_col] = joined_df[self._col].apply(encode_func)

        result_df = joined_df.drop(self._col, axis=1).rename(columns={self._target_col: self._col})
        return result_df

    def _transform_polars(self, df: PolarsDataFrame, default_value: Optional[int]) -> SparkDataFrame:
        transformed_df = df.with_columns(
            pl.col(self._col)
            .list.eval(
                pl.element().replace_strict(
                    self.get_mapping(), default=default_value if self._handle_unknown == "use_default_value" else None
                ),
                parallel=True,
            )
            .alias(self._target_col)
        )
        if self._handle_unknown == "drop":
            transformed_df = transformed_df.with_columns(pl.col(self._target_col).list.drop_nulls())
            if (
                transformed_df.with_columns(pl.col(self._target_col).list.len()).select(pl.sum(self._target_col)).item()
                == 0
            ):
                warnings.warn(
                    f"You are trying to transform dataframe with all values are unknown for {self._col}, "
                    "with `handle_unknown_strategy=drop` leads to empty dataframe",
                    LabelEncoderTransformWarning,
                )
        elif self._handle_unknown == "error":
            none_checker = transformed_df.with_columns(
                pl.col(self._target_col).list.contains(pl.lit(None, dtype=pl.Int64)).cast(pl.Int64)
            )
            if none_checker.select(pl.sum(self._target_col)).item() != 0:
                msg = f"Found unknown labels in column {self._col} during transform"
                raise ValueError(msg)

        result_df = transformed_df.drop(self._col).rename({self._target_col: self._col})
        return result_df

    def _inverse_transform_pandas(self, df: PandasDataFrame) -> PandasDataFrame:
        decoded_df = df.copy()

        def decode_func(array_col):
            return [self._inverse_mapping_list[x] for x in array_col]

        decoded_df[self._col] = decoded_df[self._col].apply(decode_func)
        return decoded_df

    def _inverse_transform_polars(self, df: PolarsDataFrame) -> PolarsDataFrame:
        mapping_size = len(self._inverse_mapping_list)
        transformed_df = df.with_columns(
            pl.col(self._col).list.eval(
                pl.element().replace_strict(old=list(range(mapping_size)), new=self._inverse_mapping_list),
                parallel=True,
            )
        )
        return transformed_df

    def _inverse_transform_spark(self, df: SparkDataFrame) -> SparkDataFrame:
        array_expr = sf.array([sf.lit(x) for x in self._inverse_mapping_list])
        decoded_df = df.withColumn(
            self._target_col, sf.transform(self._col, lambda x: sf.element_at(array_expr, x + 1))
        )
        return decoded_df.drop(self._col).withColumnRenamed(self._target_col, self._col)


class LabelEncoder:
    """
    Applies multiple label encoding rules to the data frame.

    >>> import pandas as pd
    >>> user_interactions = pd.DataFrame([
    ...     ("u1", "item_1", "item_1", [1, 2, 3]),
    ...     ("u2", "item_2", "item_2", [3, 4, 5]),
    ...     ("u3", "item_3", "item_3", [-1, -2, 4]),
    ... ], columns=["user_id", "item_1", "item_2", "list"])
    >>> user_interactions
        user_id	item_1	item_2	list
    0	u1	    item_1	item_1	[1, 2, 3]
    1	u2	    item_2	item_2	[3, 4, 5]
    2	u3	    item_3	item_3	[-1, -2, 4]
    >>> encoder = LabelEncoder([
    ...     LabelEncodingRule("user_id"),
    ...     LabelEncodingRule("item_1"),
    ...     LabelEncodingRule("item_2"),
    ...     SequenceEncodingRule("list"),
    ... ])
    >>> mapped_interactions = encoder.fit_transform(user_interactions)
    >>> mapped_interactions
       user_id  item_1  item_2  list
    0        0       0       0  [0, 1, 2]
    1        1       1       1  [2, 3, 4]
    2        2       2       2  [5, 6, 3]
    >>> encoder.mapping
    {'user_id': {'u1': 0, 'u2': 1, 'u3': 2},
    'item_1': {'item_1': 0, 'item_2': 1, 'item_3': 2},
    'item_2': {'item_1': 0, 'item_2': 1, 'item_3': 2},
    'list': {1: 0, 2: 1, 3: 2, 4: 3, 5: 4, -1: 5, -2: 6}}
    >>> encoder.inverse_mapping
    {'user_id': {0: 'u1', 1: 'u2', 2: 'u3'},
    'item_1': {0: 'item_1', 1: 'item_2', 2: 'item_3'},
    'item_2': {0: 'item_1', 1: 'item_2', 2: 'item_3'},
    'list': {0: 1, 1: 2, 2: 3, 3: 4, 4: 5, 5: -1, 6: -2}}
    >>> new_encoder = LabelEncoder([
    ...    LabelEncodingRule("user_id", encoder.mapping["user_id"]),
    ...    LabelEncodingRule("item_1", encoder.mapping["item_1"]),
    ...    LabelEncodingRule("item_2", encoder.mapping["item_2"]),
    ...    SequenceEncodingRule("list", encoder.mapping["list"]),
    ... ])
    >>> new_encoder.inverse_transform(mapped_interactions)
      user_id item_1 item_2	list
    0      u1 item_1 item_1	[1, 2, 3]
    1      u2 item_2 item_2	[3, 4, 5]
    2      u3 item_3 item_3	[-1, -2, 4]
    <BLANKLINE>
    """

    def __init__(self, rules: Sequence[BaseLabelEncodingRule]):
        """
        :param rules: Sequence of rules.
        """
        self.rules = rules

    @property
    def mapping(self) -> Mapping[str, Mapping]:
        """
        Returns mapping of each column in given rules.
        """
        return {r.column: r.get_mapping() for r in self.rules}

    @property
    def inverse_mapping(self) -> Mapping[str, Mapping]:
        """
        Returns inverse mapping of each column in given rules.
        """
        return {r.column: r.get_inverse_mapping() for r in self.rules}

    def fit(self, df: DataFrameLike) -> "LabelEncoder":
        """
        Fits an encoder by the input data frame with given rules.

        :param df: input dataframe.
        :returns: fitted LabelEncoder.
        """
        for rule in self.rules:
            rule.fit(df)
        return self

    def partial_fit(self, df: DataFrameLike) -> "LabelEncoder":
        """
        Fits an already fitted encoder by the new input data frame with given rules.
        If encoder has not been fitted yet - performs default fit.

        :param df: input dataframe.
        :returns: fitted LabelEncoder.
        """
        for rule in self.rules:
            rule.partial_fit(df)
        return self

    def transform(self, df: DataFrameLike) -> DataFrameLike:
        """
        Transforms the input data frame.
        If the input data frame contains unknown labels then they will be transformed by handle unknown strategy.

        :param df: input dataframe.
        :returns: transformed dataframe.
        """
        for rule in self.rules:
            df = rule.transform(df)
        return df

    def inverse_transform(self, df: DataFrameLike) -> DataFrameLike:
        """
        Performs inverse transform of the input data frame.

        :param df: input dataframe.
        :returns: initial dataframe.
        """
        for rule in self.rules:
            df = rule.inverse_transform(df)
        return df

    def fit_transform(self, df: DataFrameLike) -> DataFrameLike:
        """
        Fits an encoder by the input data frame with given rules and transforms the input data frame.

        :param df: input dataframe.
        :returns: transformed dataframe.
        """
        return self.fit(df).transform(df)

    def set_handle_unknowns(self, handle_unknown_rules: dict[str, HandleUnknownStrategies]) -> None:
        """
        Modify handle unknown strategy on already fitted encoder.

        :param handle_unknown_rules: handle unknown rule.

        Example: {"item_id" : None, "user_id" : -1, "category_column" : "last"}

        Default value examples:
            If ``None``, then keep null.
            If ``int`` value, then fill by that value.
            If ``str`` value, should be \"last\" only, then fill by n_classes number.
            Default ``None``.
        """
        columns = [i.column for i in self.rules]
        for column, handle_unknown in handle_unknown_rules.items():
            if column not in columns:
                msg = f"Column {column} not found."
                raise ValueError(msg)
            rule = list(filter(lambda x: x.column == column, self.rules))
            rule[0].set_handle_unknown(handle_unknown)

    def set_default_values(self, default_value_rules: dict[str, Optional[Union[int, str]]]) -> None:
        """
        Modify handle unknown strategy on already fitted encoder.
        Default value that will fill the unknown labels
        after transform if handle_unknown is set to ``use_default_value``.

        :param default_value_rules: Dictionary of default values to set for columns.

        Example: {"item_id" : "error", "user_id" : "use_default_value"}

        Default value examples:
            When set to ``error`` an error will be raised in case an unknown label is present during transform.
            When set to ``use_default_value``, the encoded value of unknown label will be set
            to the value given for the parameter default_value.
            Default: ``error``.
        """
        columns = [i.column for i in self.rules]
        for column, default_value in default_value_rules.items():
            if column not in columns:
                msg = f"Column {column} not found."
                raise ValueError(msg)
            rule = list(filter(lambda x: x.column == column, self.rules))
            rule[0].set_default_value(default_value)

    def save(
        self,
        path: str,
    ) -> None:
        encoder_dict = {}
        encoder_dict["_class_name"] = self.__class__.__name__

        base_path = Path(path).with_suffix(".replay").resolve()
        if os.path.exists(base_path):  # pragma: no cover
            msg = "There is already LabelEncoder object saved at the given path. File will be overwrited."
            warnings.warn(msg)
        else:  # pragma: no cover
            base_path.mkdir(parents=True, exist_ok=True)

        encoder_dict["rule_names"] = []

        for rule in self.rules:
            path_suffix = f"{rule.__class__.__name__}_{rule.column}"
            rule.save(str(base_path) + f"/rules/{path_suffix}")
            encoder_dict["rule_names"].append(path_suffix)

        with open(base_path / "init_args.json", "w+") as file:
            json.dump(encoder_dict, file)

    @classmethod
    def load(cls, path: str) -> "LabelEncoder":
        base_path = Path(path).with_suffix(".replay").resolve()
        with open(base_path / "init_args.json") as file:
            encoder_dict = json.loads(file.read())
        rules = []
        for root, dirs, files in os.walk(str(base_path) + "/rules/"):
            for d in dirs:
                if d.split(".")[0] in encoder_dict["rule_names"]:
                    with open(root + d + "/init_args.json") as file:
                        encoder_rule_dict = json.loads(file.read())
                    rules.append(globals()[encoder_rule_dict["_class_name"]].load(root + d))

        encoder = cls(rules=rules)
        return encoder



import collections
import logging
import os
import pickle
import warnings
from collections.abc import Iterable
from typing import Any, Optional, Union

import numpy as np
import pandas as pd
from numpy.random import default_rng

from .session_handler import State
from .types import PYSPARK_AVAILABLE, DataFrameLike, MissingImport, NumType, PolarsDataFrame, SparkDataFrame

if PYSPARK_AVAILABLE:
    import pyspark.sql.types as st
    from pyspark.ml.linalg import DenseVector, Vectors, VectorUDT
    from pyspark.sql import (
        Column,
        SparkSession,
        Window,
        functions as sf,
    )
    from pyspark.sql.column import _to_java_column, _to_seq
    from pyspark.sql.types import DoubleType, IntegerType, StructField, StructType
else:
    Column = MissingImport


class PolarsConvertToSparkWarning(Warning):
    """
    Direct PolarsDataFrame to SparkDataFrame convertation warning.
    """


class SparkCollectToMasterWarning(Warning):  # pragma: no cover
    """
    Collect to master warning for Spark DataFrames.
    """


def spark_to_pandas(
    data: SparkDataFrame, allow_collect_to_master: bool = False, from_constructor: bool = True
) -> pd.DataFrame:  # pragma: no cover
    """
    Convert Spark DataFrame to Pandas DataFrame.

    :param data: Spark DataFrame.
    :param allow_collect_to_master: Flag allowing spark to make a collection to the master node, default: ``False``.

    :returns: Converted Pandas DataFrame.
    """
    warn_msg = "Spark Data Frame is collected to master node, this may lead to OOM exception for larger dataset. "
    if from_constructor:
        _msg = "To remove this warning set allow_collect_to_master=True in the recommender constructor."
    else:
        _msg = "To remove this warning set allow_collect_to_master=True."
    warn_msg += _msg
    if not allow_collect_to_master:
        warnings.warn(
            warn_msg,
            SparkCollectToMasterWarning,
        )
    return data.toPandas()


def convert2spark(data_frame: Optional[DataFrameLike]) -> Optional[SparkDataFrame]:
    """
    Converts Pandas DataFrame to Spark DataFrame

    :param data_frame: pandas DataFrame
    :return: converted data
    """
    if data_frame is None:
        return None
    if isinstance(data_frame, SparkDataFrame):
        return data_frame

    spark = State().session
    if isinstance(data_frame, PolarsDataFrame):
        warnings.warn(
            "Direct convertation PolarsDataFrame to SparkDataFrame currently is not supported, "
            "converting to pandas first",
            PolarsConvertToSparkWarning,
        )
        return spark.createDataFrame(data_frame.to_pandas())  # TODO: remove extra convertation to pandas
    return spark.createDataFrame(data_frame)


def get_top_k(
    dataframe: SparkDataFrame,
    partition_by_col: Column,
    order_by_col: list[Column],
    k: int,
) -> SparkDataFrame:
    """
    Return top ``k`` rows for each entity in ``partition_by_col`` ordered by
    ``order_by_col``.

    >>> from replay.utils.session_handler import State
    >>> spark = State().session
    >>> log = (
    ...    spark
    ...    .createDataFrame([(1, 2, 1.), (1, 3, 1.), (1, 4, 0.5), (2, 1, 1.)])
    ...    .toDF("user_id", "item_id", "relevance")
    ... )
    >>> log.show()
    +-------+-------+---------+
    |user_id|item_id|relevance|
    +-------+-------+---------+
    |      1|      2|      1.0|
    |      1|      3|      1.0|
    |      1|      4|      0.5|
    |      2|      1|      1.0|
    +-------+-------+---------+
    <BLANKLINE>
    >>> get_top_k(dataframe=log,
    ...    partition_by_col=sf.col('user_id'),
    ...    order_by_col=[sf.col('relevance').desc(), sf.col('item_id').desc()],
    ...    k=1).orderBy('user_id').show()
    +-------+-------+---------+
    |user_id|item_id|relevance|
    +-------+-------+---------+
    |      1|      3|      1.0|
    |      2|      1|      1.0|
    +-------+-------+---------+
    <BLANKLINE>

    :param dataframe: spark dataframe to filter
    :param partition_by_col: spark column to partition by
    :param order_by_col: list of spark columns to orted by
    :param k: number of first rows for each entity in ``partition_by_col`` to return
    :return: filtered spark dataframe
    """
    return (
        dataframe.withColumn(
            "temp_rank",
            sf.row_number().over(Window.partitionBy(partition_by_col).orderBy(*order_by_col)),
        )
        .filter(sf.col("temp_rank") <= k)
        .drop("temp_rank")
    )


def get_top_k_recs(
    recs: SparkDataFrame,
    k: int,
    query_column: str = "user_idx",
    rating_column: str = "relevance",
) -> SparkDataFrame:
    """
    Get top k recommendations by `rating`.

    :param recs: recommendations DataFrame
        `[user_idx, item_idx, rating]`
    :param k: length of a recommendation list
    :param id_type: id or idx
    :return: top k recommendations `[user_id, item_id, rating]`
    """
    return get_top_k(
        dataframe=recs,
        partition_by_col=sf.col(query_column),
        order_by_col=[sf.col(rating_column).desc()],
        k=k,
    )


if PYSPARK_AVAILABLE:

    @sf.udf(returnType=st.DoubleType())
    def vector_dot(one: DenseVector, two: DenseVector) -> float:  # pragma: no cover
        """
        dot product of two column vectors

        >>> from replay.utils.session_handler import State
        >>> from pyspark.ml.linalg import Vectors
        >>> spark = State().session
        >>> input_data = (
        ...     spark.createDataFrame([(Vectors.dense([1.0, 2.0]), Vectors.dense([3.0, 4.0]))])
        ...     .toDF("one", "two")
        ... )
        >>> input_data.dtypes
        [('one', 'vector'), ('two', 'vector')]
        >>> input_data.show()
        +---------+---------+
        |      one|      two|
        +---------+---------+
        |[1.0,2.0]|[3.0,4.0]|
        +---------+---------+
        <BLANKLINE>
        >>> output_data = input_data.select(vector_dot("one", "two").alias("dot"))
        >>> output_data.schema
        StructType([StructField('dot', DoubleType(), True)])
        >>> output_data.show()
        +----+
        | dot|
        +----+
        |11.0|
        +----+
        <BLANKLINE>

        :param one: vector one
        :param two: vector two
        :returns: dot product
        """
        return float(one.dot(two))

    @sf.udf(returnType=VectorUDT())
    def vector_mult(one: Union[DenseVector, NumType], two: DenseVector) -> DenseVector:  # pragma: no cover
        """
        elementwise vector multiplication

        >>> from replay.utils.session_handler import State
        >>> from pyspark.ml.linalg import Vectors
        >>> spark = State().session
        >>> input_data = (
        ...     spark.createDataFrame([(Vectors.dense([1.0, 2.0]), Vectors.dense([3.0, 4.0]))])
        ...     .toDF("one", "two")
        ... )
        >>> input_data.dtypes
        [('one', 'vector'), ('two', 'vector')]
        >>> input_data.show()
        +---------+---------+
        |      one|      two|
        +---------+---------+
        |[1.0,2.0]|[3.0,4.0]|
        +---------+---------+
        <BLANKLINE>
        >>> output_data = input_data.select(vector_mult("one", "two").alias("mult"))
        >>> output_data.schema
        StructType([StructField('mult', VectorUDT(), True)])
        >>> output_data.show()
        +---------+
        |     mult|
        +---------+
        |[3.0,8.0]|
        +---------+
        <BLANKLINE>

        :param one: vector one
        :param two: vector two
        :returns: result
        """
        return one * two

    @sf.udf(returnType=st.ArrayType(st.DoubleType()))
    def array_mult(first: st.ArrayType, second: st.ArrayType):  # pragma: no cover
        """
        elementwise array multiplication

        >>> from replay.utils.session_handler import State
        >>> spark = State().session
        >>> input_data = (
        ...     spark.createDataFrame([([1.0, 2.0], [3.0, 4.0])])
        ...     .toDF("one", "two")
        ... )
        >>> input_data.dtypes
        [('one', 'array<double>'), ('two', 'array<double>')]
        >>> input_data.show()
        +----------+----------+
        |       one|       two|
        +----------+----------+
        |[1.0, 2.0]|[3.0, 4.0]|
        +----------+----------+
        <BLANKLINE>
        >>> output_data = input_data.select(array_mult("one", "two").alias("mult"))
        >>> output_data.schema
        StructType([StructField('mult', ArrayType(DoubleType(), True), True)])
        >>> output_data.show()
        +----------+
        |      mult|
        +----------+
        |[3.0, 8.0]|
        +----------+
        <BLANKLINE>

        :param first: first array
        :param second: second array
        :returns: result
        """

        return [first[i] * second[i] for i in range(len(first))]


def multiply_scala_udf(scalar, vector):
    """
    Multiplies a scalar by a vector

    :param scalar: column with scalars
    :param vector: column with vectors
    :return: column expression
    """
    sc = SparkSession.getActiveSession().sparkContext
    _f = sc._jvm.org.apache.spark.replay.utils.ScalaPySparkUDFs.multiplyUDF()
    return Column(_f.apply(_to_seq(sc, [scalar, vector], _to_java_column)))


def get_log_info(log: SparkDataFrame, user_col="user_idx", item_col="item_idx") -> str:
    """
    Basic log statistics

    >>> from replay.utils.session_handler import State
    >>> spark = State().session
    >>> log = spark.createDataFrame([(1, 2), (3, 4), (5, 2)]).toDF("user_idx", "item_idx")
    >>> log.show()
    +--------+--------+
    |user_idx|item_idx|
    +--------+--------+
    |       1|       2|
    |       3|       4|
    |       5|       2|
    +--------+--------+
    <BLANKLINE>
    >>> get_log_info(log)
    'total lines: 3, total users: 3, total items: 2'

    :param log: interaction log containing ``user_idx`` and ``item_idx``
    :param user_col: name of a columns containing users' identificators
    :param item_col: name of a columns containing items' identificators

    :returns: statistics string
    """
    cnt = log.count()
    user_cnt = log.select(user_col).distinct().count()
    item_cnt = log.select(item_col).distinct().count()
    return ", ".join(
        [
            f"total lines: {cnt}",
            f"total users: {user_cnt}",
            f"total items: {item_cnt}",
        ]
    )


def get_stats(log: SparkDataFrame, group_by: str = "user_id", target_column: str = "relevance") -> SparkDataFrame:
    """
    Calculate log statistics: min, max, mean, median ratings, number of ratings.
    >>> from replay.utils.session_handler import get_spark_session, State
    >>> spark = get_spark_session(1, 1)
    >>> test_df = (spark.
    ...   createDataFrame([(1, 2, 1), (1, 3, 3), (1, 1, 2), (2, 3, 2)])
    ...   .toDF("user_id", "item_id", "rel")
    ...   )
    >>> get_stats(test_df, target_column='rel').show()
    +-------+--------+-------+-------+---------+----------+
    |user_id|mean_rel|max_rel|min_rel|count_rel|median_rel|
    +-------+--------+-------+-------+---------+----------+
    |      1|     2.0|      3|      1|        3|         2|
    |      2|     2.0|      2|      2|        1|         2|
    +-------+--------+-------+-------+---------+----------+
    <BLANKLINE>
    >>> get_stats(test_df, group_by='item_id', target_column='rel').show()
    +-------+--------+-------+-------+---------+----------+
    |item_id|mean_rel|max_rel|min_rel|count_rel|median_rel|
    +-------+--------+-------+-------+---------+----------+
    |      2|     1.0|      1|      1|        1|         1|
    |      3|     2.5|      3|      2|        2|         2|
    |      1|     2.0|      2|      2|        1|         2|
    +-------+--------+-------+-------+---------+----------+
    <BLANKLINE>

    :param log: spark DataFrame with ``user_id``, ``item_id`` and ``relevance`` columns
    :param group_by: column to group data by, ``user_id`` or ``item_id``
    :param target_column: column with interaction ratings
    :return: spark DataFrame with statistics
    """
    agg_functions = {
        "mean": sf.avg,
        "max": sf.max,
        "min": sf.min,
        "count": sf.count,
    }
    agg_functions_list = [
        func(target_column).alias(str(name + "_" + target_column)) for name, func in agg_functions.items()
    ]
    agg_functions_list.append(sf.expr(f"percentile_approx({target_column}, 0.5)").alias("median_" + target_column))

    return log.groupBy(group_by).agg(*agg_functions_list)


def check_numeric(feature_table: SparkDataFrame) -> None:
    """
    Check if spark DataFrame columns are of NumericType
    :param feature_table: spark DataFrame
    """
    for column in feature_table.columns:
        if not isinstance(feature_table.schema[column].dataType, st.NumericType):
            msg = f"Column {column} has type {feature_table.schema[column].dataType}, that is not numeric."
            raise ValueError(msg)


def horizontal_explode(
    data_frame: SparkDataFrame,
    column_to_explode: str,
    prefix: str,
    other_columns: list[Column],
) -> SparkDataFrame:
    """
    Transform a column with an array of values into separate columns.
    Each array must contain the same amount of values.

    >>> from replay.utils.session_handler import State
    >>> spark = State().session
    >>> input_data = (
    ...     spark.createDataFrame([(5, [1.0, 2.0]), (6, [3.0, 4.0])])
    ...     .toDF("id_col", "array_col")
    ... )
    >>> input_data.show()
    +------+----------+
    |id_col| array_col|
    +------+----------+
    |     5|[1.0, 2.0]|
    |     6|[3.0, 4.0]|
    +------+----------+
    <BLANKLINE>
    >>> horizontal_explode(input_data, "array_col", "element", [sf.col("id_col")]).show()
    +------+---------+---------+
    |id_col|element_0|element_1|
    +------+---------+---------+
    |     5|      1.0|      2.0|
    |     6|      3.0|      4.0|
    +------+---------+---------+
    <BLANKLINE>

    :param data_frame: input DataFrame
    :param column_to_explode: column with type ``array``
    :param prefix: prefix used for new columns, suffix is an integer
    :param other_columns: columns to select beside newly created
    :returns: DataFrame with elements from ``column_to_explode``
    """
    num_columns = len(data_frame.select(column_to_explode).head()[0])
    return data_frame.select(
        *other_columns,
        *[sf.element_at(column_to_explode, i + 1).alias(f"{prefix}_{i}") for i in range(num_columns)],
    )


def join_or_return(first, second, on, how):
    """
    Safe wrapper for join of two DataFrames if ``second`` parameter is None it returns ``first``.

    :param first: Spark DataFrame
    :param second: Spark DataFrame
    :param on: name of the join column
    :param how: type of join
    :return: Spark DataFrame
    """
    if second is None:
        return first
    return first.join(second, on=on, how=how)


def fallback(
    base: SparkDataFrame,
    fill: SparkDataFrame,
    k: int,
    query_column: str = "user_idx",
    item_column: str = "item_idx",
    rating_column: str = "relevance",
) -> SparkDataFrame:
    """
    Fill missing recommendations for users that have less than ``k`` recommended items.
    Score values for the fallback model may be decreased to preserve sorting.

    :param base: base recommendations that need to be completed
    :param fill: extra recommendations
    :param k: desired recommendation list lengths for each user
    :param id_type: id or idx
    :return: augmented recommendations
    """
    if fill is None:
        return base
    if base.count() == 0:
        return get_top_k_recs(fill, k, query_column=query_column, rating_column=rating_column)
    margin = 0.1
    min_in_base = base.agg({rating_column: "min"}).first()[0]
    max_in_fill = fill.agg({rating_column: "max"}).first()[0]
    diff = max_in_fill - min_in_base
    fill = fill.withColumnRenamed(rating_column, "relevance_fallback")
    if diff >= 0:
        fill = fill.withColumn("relevance_fallback", sf.col("relevance_fallback") - diff - margin)
    recs = base.join(fill, on=[query_column, item_column], how="full_outer")
    recs = recs.withColumn(rating_column, sf.coalesce(rating_column, "relevance_fallback")).select(
        query_column, item_column, rating_column
    )
    recs = get_top_k_recs(recs, k, query_column=query_column, rating_column=rating_column)
    return recs


def cache_if_exists(dataframe: Optional[SparkDataFrame]) -> Optional[SparkDataFrame]:
    """
    Cache a DataFrame
    :param dataframe: Spark DataFrame or None
    :return: DataFrame or None
    """
    if dataframe is not None:
        return dataframe.cache()
    return dataframe


def unpersist_if_exists(dataframe: Optional[SparkDataFrame]) -> None:
    """
    :param dataframe: DataFrame or None
    """
    if dataframe is not None and dataframe.is_cached:
        dataframe.unpersist()


def join_with_col_renaming(
    left: SparkDataFrame,
    right: SparkDataFrame,
    on_col_name: Union[str, list],
    how: str = "inner",
    suffix="join",
) -> SparkDataFrame:
    """
    There is a bug in some Spark versions (e.g. 3.0.2), which causes errors
    in joins of DataFrames derived form the same DataFrame on the columns with the same name:
    https://issues.apache.org/jira/browse/SPARK-14948
    https://issues.apache.org/jira/browse/SPARK-36815.

    The function renames columns stated in `on_col_name` in one dataframe,
    performs join and removes renamed columns.

    :param left: left-side dataframe
    :param right: right-side dataframe
    :param on_col_name: column names to join on
    :param how: join type
    :param suffix: suffix added to `on_col_name` values to name temporary column
    :return: join result
    """
    if isinstance(on_col_name, str):
        on_col_name = [on_col_name]

    on_condition = sf.lit(True)
    for name in on_col_name:
        if how == "right":
            left = left.withColumnRenamed(name, f"{name}_{suffix}")
        else:
            right = right.withColumnRenamed(name, f"{name}_{suffix}")
        on_condition &= sf.col(name) == sf.col(f"{name}_{suffix}")

    return (left.join(right, on=on_condition, how=how)).drop(*[f"{name}_{suffix}" for name in on_col_name])


def process_timestamp_column(
    dataframe: SparkDataFrame,
    column_name: str,
    date_format: Optional[str] = None,
) -> SparkDataFrame:
    """
    Convert ``column_name`` column of numeric/string/timestamp type
    to TimestampType.
    Return original ``dataframe`` if the column has TimestampType.
    Treats numbers as unix timestamp, treats strings as
    a string representation of dates in ``date_format``.
    Date format is inferred by pyspark if not defined by ``date_format``.

    :param dataframe: spark dataframe
    :param column_name: name of ``dataframe`` column to convert
    :param date_format: datetime pattern passed to
        ``to_timestamp`` pyspark sql function
    :return: dataframe with updated column ``column_name``
    """
    if column_name not in dataframe.columns:
        msg = f"Column {column_name} not found"
        raise ValueError(msg)

    # no conversion needed
    if isinstance(dataframe.schema[column_name].dataType, st.TimestampType):
        return dataframe

    # unix timestamp
    if isinstance(dataframe.schema[column_name].dataType, st.NumericType):
        return dataframe.withColumn(column_name, sf.to_timestamp(sf.from_unixtime(sf.col(column_name))))

    # datetime in string format
    dataframe = dataframe.withColumn(
        column_name,
        sf.to_timestamp(sf.col(column_name), format=date_format),
    )
    return dataframe


if PYSPARK_AVAILABLE:

    @sf.udf(returnType=VectorUDT())
    def list_to_vector_udf(array: st.ArrayType) -> DenseVector:  # pragma: no cover
        """
        convert spark array to vector

        :param array: spark Array to convert
        :return:  spark DenseVector
        """
        return Vectors.dense(array)

    @sf.udf(returnType=st.FloatType())
    def vector_squared_distance(first: DenseVector, second: DenseVector) -> float:  # pragma: no cover
        """
        :param first: first vector
        :param second: second vector
        :returns: squared distance value
        """
        return float(first.squared_distance(second))

    @sf.udf(returnType=st.FloatType())
    def vector_euclidean_distance_similarity(first: DenseVector, second: DenseVector) -> float:  # pragma: no cover
        """
        :param first: first vector
        :param second: second vector
        :returns: 1/(1 + euclidean distance value)
        """
        return 1 / (1 + float(first.squared_distance(second)) ** 0.5)

    @sf.udf(returnType=st.FloatType())
    def cosine_similarity(first: DenseVector, second: DenseVector) -> float:  # pragma: no cover
        """
        :param first: first vector
        :param second: second vector
        :returns: cosine similarity value
        """
        num = first.dot(second)
        denom = first.dot(first) ** 0.5 * second.dot(second) ** 0.5
        return float(num / denom)


def cache_temp_view(df: SparkDataFrame, name: str) -> None:
    """
    Create Spark SQL temporary view with `name` and cache it
    """
    spark = State().session
    df.createOrReplaceTempView(name)
    spark.sql(f"cache table {name}")


def drop_temp_view(temp_view_name: str) -> None:
    """
    Uncache and drop Spark SQL temporary view
    """
    spark = State().session
    spark.catalog.dropTempView(temp_view_name)


def sample_top_k_recs(pairs: SparkDataFrame, k: int, seed: Optional[int] = None):
    """
    Sample k items for each user with probability proportional to the relevance score.

    Motivation: sometimes we have a pre-defined list of items for each user
    and could use `predict_pairs` method of RePlay models to score them.
    After that we could select top K most relevant items for each user
    with `replay.utils.spark_utils.get_top_k_recs` or sample them with
    probabilities proportional to their relevance score
    with `replay.utils.spark_utils.sample_top_k_recs` to get more diverse recommendations.

    :param pairs: spark dataframe with columns ``[user_idx, item_idx, relevance]``
    :param k: number of items for each user to return
    :param seed: random seed
    :return:  spark dataframe with columns ``[user_idx, item_idx, relevance]``
    """
    pairs = pairs.withColumn(
        "probability",
        sf.col("relevance") / sf.sum("relevance").over(Window.partitionBy("user_idx")),
    )

    def grouped_map(pandas_df: pd.DataFrame) -> pd.DataFrame:  # pragma: no cover
        user_idx = pandas_df["user_idx"][0]

        local_rng = default_rng(seed + user_idx) if seed is not None else default_rng()

        items_positions = local_rng.choice(
            np.arange(pandas_df.shape[0]),
            size=min(k, pandas_df.shape[0]),
            p=pandas_df["probability"].values,
            replace=False,
        )

        return pd.DataFrame(
            {
                "user_idx": k * [user_idx],
                "item_idx": pandas_df["item_idx"].values[items_positions],
                "relevance": pandas_df["relevance"].values[items_positions],
            }
        )

    rec_schema = StructType(
        [
            StructField("user_idx", IntegerType()),
            StructField("item_idx", IntegerType()),
            StructField("relevance", DoubleType()),
        ]
    )
    recs = pairs.groupby("user_idx").applyInPandas(grouped_map, rec_schema)

    return recs


def filter_cold(
    df: Optional[SparkDataFrame],
    warm_df: SparkDataFrame,
    col_name: str,
) -> tuple[int, Optional[SparkDataFrame]]:
    """
    Filter out new user/item ids absent in `warm_df`.
    Return number of new users/items and filtered dataframe.

    :param df: spark dataframe with columns ``[`col_name`, ...]``
    :param warm_df: spark dataframe with column ``[`col_name`]``,
        containing ids of `warm` users/items
    :param col_name: name of a column
    :return:  filtered spark dataframe columns ``[`col_name`, ...]``
    """
    if df is None:
        return 0, df

    num_cold = df.select(col_name).distinct().join(warm_df, on=col_name, how="anti").count()

    if num_cold == 0:
        return 0, df

    return num_cold, df.join(warm_df.select(col_name), on=col_name, how="inner")


def get_unique_entities(
    df: Union[Iterable, SparkDataFrame],
    column: str,
) -> SparkDataFrame:
    """
    Get unique values from ``df`` and put them into dataframe with column ``column``.
    :param df: spark dataframe with ``column`` or python iterable
    :param column: name of a column
    :return:  spark dataframe with column ``[`column`]``
    """
    spark = State().session
    if isinstance(df, SparkDataFrame):
        unique = df.select(column).distinct()
    elif isinstance(df, collections.abc.Iterable):
        unique = spark.createDataFrame(data=pd.DataFrame(pd.unique(list(df)), columns=[column]))
    else:
        msg = f"Wrong type {type(df)}"
        raise ValueError(msg)
    return unique


def return_recs(recs: SparkDataFrame, recs_file_path: Optional[str] = None) -> Optional[SparkDataFrame]:
    """
    Save dataframe `recs` to `recs_file_path` if presents otherwise cache
    and materialize the dataframe.

    :param recs: dataframe with recommendations
    :param recs_file_path: absolute path to save recommendations as a parquet file.
    :return: cached and materialized dataframe `recs` if `recs_file_path` is provided otherwise None
    """
    if recs_file_path is None:
        output = recs.cache()
        output.count()
        return output

    recs.write.parquet(path=recs_file_path, mode="overwrite")
    return None


def save_picklable_to_parquet(obj: Any, path: str) -> None:
    """
    Function dumps object to disk or hdfs in parquet format.

    :param obj: object to be saved
    :param path: path to dump
    :return:
    """

    sc = State().session.sparkContext
    # We can use `RDD.saveAsPickleFile`, but it has no "overwrite" parameter
    pickled_instance = pickle.dumps(obj)
    Record = collections.namedtuple("Record", ["data"])  # noqa: PYI024
    rdd = sc.parallelize([Record(pickled_instance)])
    instance_df = rdd.map(lambda rec: Record(bytearray(rec.data))).toDF()
    instance_df.write.mode("overwrite").parquet(path)


def load_pickled_from_parquet(path: str) -> Any:
    """
    Function loads object from disk or hdfs,
    what was dumped via `save_picklable_to_parquet` function.

    :param path: source path
    :return: unpickled object
    """
    spark = State().session
    df = spark.read.parquet(path)
    pickled_instance = df.rdd.map(lambda row: bytes(row.data)).first()
    return pickle.loads(pickled_instance)


def assert_omp_single_thread():
    """
    Check that OMP_NUM_THREADS is set to 1 and warn if not.

    PyTorch uses multithreading for cpu math operations via OpenMP library. Sometimes this
    leads to failures when OpenMP multithreading is mixed with multiprocessing.
    """
    omp_num_threads = os.environ.get("OMP_NUM_THREADS", None)
    if omp_num_threads != "1":
        msg = (
            f'Environment variable "OMP_NUM_THREADS" is set to "{omp_num_threads}". '
            f"Set it to 1 if the working process freezes."
        )
        logging.getLogger("replay").warning(msg)



from typing import Optional

import numpy as np
import pandas as pd

from replay.utils import PYSPARK_AVAILABLE, DataFrameLike, PandasDataFrame, SparkDataFrame

if PYSPARK_AVAILABLE:
    import pyspark.sql.functions as sf
    from pyspark.sql.window import Window


class Sessionizer:
    """
    Create and filter sessions from given interactions.
    Session ids are formed as subtraction between unique
    users cumulative sum and number of entries inside each
    user history that are greater than session gap.

    >>> import pandas as pd
    >>> time_interactions = pd.DataFrame({
    ...    "user_id": [1, 1, 1, 2, 2, 2, 3, 3, 3, 3],
    ...    "item_id": [3, 7, 10, 5, 8, 11, 4, 9, 2, 5],
    ...    "timestamp": [1, 2, 3, 3, 2, 1, 3, 12, 1, 4]
    ... })
    >>> time_interactions
       user_id  item_id  timestamp
    0        1        3          1
    1        1        7          2
    2        1       10          3
    3        2        5          3
    4        2        8          2
    5        2       11          1
    6        3        4          3
    7        3        9         12
    8        3        2          1
    9        3        5          4
    >>> Sessionizer(session_gap=5).transform(time_interactions)
       user_id  item_id  timestamp  session_id
    0        1        3          1           2
    1        1        7          2           2
    2        1       10          3           2
    3        2        5          3           5
    4        2        8          2           5
    5        2       11          1           5
    6        3        4          3           9
    7        3        9         12           8
    8        3        2          1           9
    9        3        5          4           9
    <BLANKLINE>
    """

    def __init__(
        self,
        user_column: str = "user_id",
        time_column: str = "timestamp",
        session_column: str = "session_id",
        session_gap: int = 86400,
        time_column_format: str = "yyyy-MM-dd HH:mm:ss",
        min_inter_per_session: Optional[int] = None,
        max_inter_per_session: Optional[int] = None,
        min_sessions_per_user: Optional[int] = None,
        max_sessions_per_user: Optional[int] = None,
    ):
        """
        :param user_column: Name of user interaction column,
            default: ``user_id``.
        :param time_column: Name of time column,
            default: ``timestamp``.
        :param session_column: Name of column with resulting sessions,
            default: ``session_id``.
        :param session_gap: The interval between interactions,
            beyond which the division into different sessions takes place.
            If time_column has `date` type, then it will be converted
            in `unix_timestamp` type using parameter `time_column_format`.
            default: 86400.
        :param time_column_format: Format of time_column,
            needs for convert `time_column` into `unix_timestamp` type.
            If `time_column` has already transformed into unix_timestamp type,
            then you can omit this parameter.
            default: ``yyyy-MM-dd HH:mm:ss``
        :param min_inter_per_session: Minimum number of interactions per session.
            Sessions with less number of interactions are ignored.
            If None, filter doesn't apply,
            Default: ``None``.
        :param max_inter_per_session: Maximum number of interactions per session.
            Sessions with greater number of interactions are ignored.
            If None, filter doesn't apply. Must be less than `min_inter_per_session`.
            Default: ``None``.
        :param min_sessions_per_user: Minimum number of sessions per user.
            Users with less number of sessions are ignored.
            If None, filter doesn't apply.
            Default: ``None``.
        :param max_sessions_per_user: Maximum number of sessions per user.
            Users with greater number of sessions are ignored.
            If None, filter doesn't apply. Must be less
            than `min_sessions_per_user`. Default: ``None``.
        """
        self.user_column = user_column
        self.time_column = time_column
        self.session_column = session_column
        self.session_gap = session_gap
        self.time_column_format = time_column_format
        self.min_inter_per_session = min_inter_per_session
        self.max_inter_per_session = max_inter_per_session
        self.min_sessions_per_user = min_sessions_per_user
        self.max_sessions_per_user = max_sessions_per_user
        self._sanity_check()

    def _sanity_check(self) -> None:
        if self.min_inter_per_session:
            assert self.min_inter_per_session > 0
        if self.min_sessions_per_user:
            assert self.min_sessions_per_user > 0
        if self.min_inter_per_session and self.max_inter_per_session:
            assert self.min_inter_per_session <= self.max_inter_per_session
        if self.min_sessions_per_user and self.max_sessions_per_user:
            assert self.min_sessions_per_user <= self.max_sessions_per_user

    def _to_unix_timestamp(self, interactions: DataFrameLike) -> DataFrameLike:
        if isinstance(interactions, SparkDataFrame):
            return self._to_unix_timestamp_spark(interactions)

        return self._to_unix_timestamp_pandas(interactions)

    def _to_unix_timestamp_pandas(self, interactions: PandasDataFrame) -> PandasDataFrame:
        time_column_type = dict(interactions.dtypes)[self.time_column]
        if time_column_type == np.dtype("datetime64[ns]"):
            interactions = interactions.copy(deep=True)
            interactions[self.time_column] = (
                interactions[self.time_column] - pd.Timestamp("1970-01-01")
            ) // pd.Timedelta("1s")

        return interactions

    def _to_unix_timestamp_spark(self, interactions: SparkDataFrame) -> SparkDataFrame:
        time_column_type = dict(interactions.dtypes)[self.time_column]
        if time_column_type == "date":
            interactions = interactions.withColumn(
                self.time_column, sf.unix_timestamp(self.time_column, self.time_column_format)
            )

        return interactions

    def _create_sessions(self, data: DataFrameLike) -> DataFrameLike:
        if isinstance(data, SparkDataFrame):
            return self._create_sessions_spark(data)

        return self._create_sessions_pandas(data)

    def _create_sessions_pandas(self, data: PandasDataFrame) -> PandasDataFrame:
        res = data.copy(deep=True)

        diff = res[self.time_column] - res.sort_values([self.user_column, self.time_column]).groupby(self.user_column)[
            self.time_column
        ].shift(1)
        nan_mask = diff.isna()
        diff = diff >= self.session_gap
        diff[nan_mask] = True
        res["timestamp_diff"] = diff
        res["cumsum_timestamp_diff"] = (
            res.sort_values([self.user_column, self.time_column, "timestamp_diff"], ascending=[True, True, False])
            .groupby(self.user_column, sort=False)["timestamp_diff"]
            .cumsum()
        )

        user_count = res.groupby(self.user_column)[self.user_column].count().cumsum().to_frame()
        user_count.rename(columns={self.user_column: "count"}, inplace=True)
        res = res.join(user_count, how="left", on=self.user_column)
        res[self.session_column] = res["count"] - res["cumsum_timestamp_diff"]

        res.drop(columns=["timestamp_diff", "cumsum_timestamp_diff", "count"], inplace=True)

        return res

    def _create_sessions_spark(self, data: SparkDataFrame) -> SparkDataFrame:
        data_with_diff = data.withColumn(
            "timestamp_diff",
            sf.col(self.time_column)
            - sf.lag(self.time_column, 1).over(Window.partitionBy(self.user_column).orderBy(self.time_column))
            >= self.session_gap,
        )
        data_with_diff = data_with_diff.na.fill(True).withColumn(
            "timestamp_diff", sf.col("timestamp_diff").cast("long")
        )
        data_with_sum_timediff = data_with_diff.withColumn(
            "cumsum_timestamp_diff",
            sf.sum("timestamp_diff").over(
                Window.partitionBy(self.user_column).orderBy(sf.col(self.time_column), sf.col("timestamp_diff").desc())
            ),
        )

        grouped_users = data_with_sum_timediff.groupBy(self.user_column).count()
        grouped_users_with_cumsum = grouped_users.withColumn(
            "cumsum_user_count",
            sf.sum("count").over(Window.partitionBy(sf.lit(0)).orderBy(self.user_column)),
        ).drop("count")

        result = (
            data_with_sum_timediff.join(grouped_users_with_cumsum, self.user_column, "left")
            .withColumn(
                self.session_column,
                sf.col("cumsum_user_count") - sf.col("cumsum_timestamp_diff"),
            )
            .drop(
                "timestamp_diff",
                "cumsum_timestamp_diff",
                "cumsum_user_count",
            )
        )

        return result

    def _filter_sessions(self, interactions: DataFrameLike) -> DataFrameLike:
        if isinstance(interactions, SparkDataFrame):
            return self._filter_sessions_spark(interactions)

        return self._filter_sessions_pandas(interactions)

    def _filter_sessions_pandas(self, interactions: PandasDataFrame) -> PandasDataFrame:
        interactions["count"] = interactions.groupby(self.session_column, sort=False)[self.session_column].transform(
            "count"
        )
        if self.min_inter_per_session:
            interactions = interactions[interactions["count"] >= self.min_inter_per_session]
        if self.max_inter_per_session:
            interactions = interactions[interactions["count"] <= self.max_inter_per_session]

        interactions["nunique"] = interactions.groupby(self.user_column, sort=False)[self.session_column].transform(
            "nunique"
        )
        if self.min_sessions_per_user:
            interactions = interactions[interactions["nunique"] >= self.min_sessions_per_user]
        if self.max_sessions_per_user:
            interactions = interactions[interactions["nunique"] <= self.max_sessions_per_user]

        interactions.drop(columns=["count", "nunique"], inplace=True)

        return interactions

    def _filter_sessions_spark(self, interactions: SparkDataFrame) -> SparkDataFrame:
        entries_counter = interactions.groupby(self.session_column).count()
        if self.min_inter_per_session:
            entries_counter = entries_counter.filter(sf.col("count") >= self.min_inter_per_session)
        if self.max_inter_per_session:
            entries_counter = entries_counter.filter(sf.col("count") <= self.max_inter_per_session)

        filtered_interactions = interactions.join(
            entries_counter.select(self.session_column), self.session_column, how="right"
        )

        nunique = filtered_interactions.groupby(self.user_column).agg(
            sf.expr("count(distinct session_id)").alias("nunique")
        )
        if self.min_sessions_per_user:
            nunique = nunique.filter(sf.col("nunique") >= self.min_sessions_per_user)
        if self.max_sessions_per_user:
            nunique = nunique.filter(sf.col("nunique") <= self.max_sessions_per_user)

        result = filtered_interactions.join(
            nunique.select(self.user_column),
            self.user_column,
            how="right",
        )
        return result

    def transform(self, interactions: DataFrameLike) -> DataFrameLike:
        r"""Create and filter sessions from given interactions.

        :param interactions: DataFrame containing columns ``user_column``, ``time_column``.

        :returns: DataFrame with created and filtered sessions.
        """
        columns_order = list(interactions.columns)
        interactions = self._to_unix_timestamp(interactions)
        result = self._create_sessions(interactions)
        result = self._filter_sessions(result)
        columns_order += [self.session_column]

        result = result.select(*columns_order) if isinstance(result, SparkDataFrame) else result[columns_order]

        return result



"""
``Dataset`` universal dataset class for manipulating interactions and feed data to models.
"""

from __future__ import annotations

import json
from collections.abc import Iterable, Sequence
from pathlib import Path
from typing import Callable, Optional, Union

import numpy as np
from pandas import read_parquet as pd_read_parquet
from polars import read_parquet as pl_read_parquet

from replay.utils import (
    PYSPARK_AVAILABLE,
    DataFrameLike,
    PandasDataFrame,
    PolarsDataFrame,
    SparkDataFrame,
)
from replay.utils.session_handler import get_spark_session

from .schema import FeatureHint, FeatureInfo, FeatureSchema, FeatureSource, FeatureType

if PYSPARK_AVAILABLE:
    import pyspark.sql.functions as sf
    from pyspark.storagelevel import StorageLevel


class Dataset:
    """
    Universal dataset for feeding data to models.
    """

    def __init__(
        self,
        feature_schema: FeatureSchema,
        interactions: DataFrameLike,
        query_features: Optional[DataFrameLike] = None,
        item_features: Optional[DataFrameLike] = None,
        check_consistency: bool = True,
        categorical_encoded: bool = False,
    ):
        """
        :param feature_schema: mapping of columns names and feature infos.
        :param interactions: dataframe with interactions.
        :param query_features: dataframe with query features,
            defaults: ```None```.
        :param item_features: dataframe with item features,
            defaults: ```None```.
        :param check_consistency: the parameter responsible for checking the consistency of the data,
            defaults: ```True```.
        :param categorical_encoded: the parameter responsible for checking the categorical features
            encoded validity,
            defaults: ```False```.
        """
        self._interactions = interactions
        self._query_features = query_features
        self._item_features = item_features

        self._assign_df_type()

        self._categorical_encoded = categorical_encoded

        try:
            feature_schema.item_id_column
        except Exception as exception:
            msg = "Item id column is not set."
            raise ValueError(msg) from exception

        try:
            feature_schema.query_id_column
        except Exception as exception:
            msg = "Query id column is not set."
            raise ValueError(msg) from exception

        if self.item_features is not None and not check_dataframes_types_equal(self._interactions, self.item_features):
            msg = "Interactions and item features should have the same type."
            raise TypeError(msg)
        if self.query_features is not None and not check_dataframes_types_equal(
            self._interactions, self.query_features
        ):
            msg = "Interactions and query features should have the same type."
            raise TypeError(msg)

        self._get_feature_source_map()
        self._get_ids_source_map()

        self._feature_schema = self._fill_feature_schema(feature_schema)

        if check_consistency:
            if self.query_features is not None:
                self._check_ids_consistency(hint=FeatureHint.QUERY_ID)
            if self.item_features is not None:
                self._check_ids_consistency(hint=FeatureHint.ITEM_ID)
            if self._categorical_encoded:
                self._check_encoded()

    @property
    def is_categorical_encoded(self) -> bool:
        """
        :returns: is categorical features are encoded.
        """
        return self._categorical_encoded

    @property
    def interactions(self) -> DataFrameLike:
        """
        :returns: interactions dataset.
        """
        return self._interactions

    @property
    def query_features(self) -> Optional[DataFrameLike]:
        """
        :returns: query features dataset.
        """
        return self._query_features

    @property
    def item_features(self) -> Optional[DataFrameLike]:
        """
        :returns: item features dataset.
        """
        return self._item_features

    @property
    def query_ids(self) -> DataFrameLike:
        """
        :returns: dataset with unique query ids.
        """
        query_column_df = self._ids_feature_map[FeatureHint.QUERY_ID]
        if self.is_pandas:
            assert isinstance(query_column_df, PandasDataFrame)
            query_ids = PandasDataFrame(
                {self.feature_schema.query_id_column: query_column_df[self.feature_schema.query_id_column].unique()}
            )
        if self.is_spark:
            assert isinstance(query_column_df, SparkDataFrame)
            query_ids = query_column_df.select(self.feature_schema.query_id_column).distinct()
        if self.is_polars:
            assert isinstance(query_column_df, PolarsDataFrame)
            query_ids = query_column_df.select(self.feature_schema.query_id_column).unique()

        return query_ids

    @property
    def item_ids(self) -> DataFrameLike:
        """
        :returns: dataset with unique item ids.
        """
        item_column_df = self._ids_feature_map[FeatureHint.ITEM_ID]
        if self.is_pandas:
            assert isinstance(item_column_df, PandasDataFrame)
            all_item_ids = item_column_df[self.feature_schema.item_id_column]
            unique_item_ids = all_item_ids.unique()
            item_ids = PandasDataFrame({self.feature_schema.item_id_column: unique_item_ids})
        if self.is_spark:
            assert isinstance(item_column_df, SparkDataFrame)
            item_ids = item_column_df.select(self.feature_schema.item_id_column).distinct()
        if self.is_polars:
            assert isinstance(item_column_df, PolarsDataFrame)
            item_ids = item_column_df.select(self.feature_schema.item_id_column).unique()

        return item_ids

    @property
    def query_count(self) -> int:
        """
        :returns: the number of queries.
        """
        query_count = self.feature_schema.query_id_feature.cardinality
        assert query_count is not None
        return query_count

    @property
    def item_count(self) -> int:
        """
        :returns: The number of items.
        """
        item_count = self.feature_schema.item_id_feature.cardinality
        assert item_count is not None
        return item_count

    @property
    def feature_schema(self) -> FeatureSchema:
        """
        :returns: List of features.
        """
        return self._feature_schema

    def _get_df_type(self) -> str:
        """
        :returns: Stored dataframe type.
        """
        if self.is_spark:
            return "spark"
        if self.is_pandas:
            return "pandas"
        if self.is_polars:
            return "polars"
        msg = "No known dataframe types are provided"
        raise ValueError(msg)

    def _to_parquet(self, df: DataFrameLike, path: Path) -> None:
        """
        Save the content of the dataframe in parquet format to the provided path.

        :param df: Dataframe to save.
        :param path: Path to save the dataframe to.
        """
        if self.is_spark:
            path = str(path)
            df = df.withColumn("idx", sf.monotonically_increasing_id())
            df.write.mode("overwrite").parquet(path)
        elif self.is_pandas:
            df.to_parquet(path)
        elif self.is_polars:
            df.write_parquet(path)
        else:
            msg = """
            _to_parquet() can only be used to save polars|pandas|spark dataframes;
            No known dataframe types are provided
            """
            raise TypeError(msg)

    @staticmethod
    def _read_parquet(path: Path, mode: str) -> Union[SparkDataFrame, PandasDataFrame, PolarsDataFrame]:
        """
        Read the parquet file as dataframe.

        :param path: The parquet file path.
        :param mode: Dataframe type. Can be spark|pandas|polars.
        :returns: The dataframe read from the file.
        """
        if mode == "spark":
            path = str(path)
            spark_session = get_spark_session()
            df = spark_session.read.parquet(path)
            if "idx" in df.columns:
                df = df.orderBy("idx").drop("idx")
            return df
        if mode == "pandas":
            df = pd_read_parquet(path)
            if "idx" in df.columns:
                df = df.set_index("idx").reset_index(drop=True)
            return df
        if mode == "polars":
            df = pl_read_parquet(path, use_pyarrow=True)
            if "idx" in df.columns:
                df = df.sort("idx").drop("idx")
            return df
        msg = f"_read_parquet() can only be used to read polars|pandas|spark dataframes, not {mode}"
        raise TypeError(msg)

    def save(self, path: str) -> None:
        """
        Save the Dataset to the provided path.

        :param path: Path to save the Dataset to.
        """
        dataset_dict = {}
        dataset_dict["_class_name"] = self.__class__.__name__

        interactions_type = self._get_df_type()
        dataset_dict["init_args"] = {
            "feature_schema": [],
            "interactions": interactions_type,
            "item_features": (interactions_type if self.item_features is not None else None),
            "query_features": (interactions_type if self.query_features is not None else None),
            "check_consistency": False,
            "categorical_encoded": self._categorical_encoded,
        }

        for feature in self.feature_schema.all_features:
            dataset_dict["init_args"]["feature_schema"].append(
                {
                    "column": feature.column,
                    "feature_type": feature.feature_type.name,
                    "feature_hint": (feature.feature_hint.name if feature.feature_hint else None),
                }
            )

        base_path = Path(path).with_suffix(".replay").resolve()
        base_path.mkdir(parents=True, exist_ok=True)

        with open(base_path / "init_args.json", "w+") as file:
            json.dump(dataset_dict, file)

        df_data = {
            "interactions": self.interactions,
            "item_features": self.item_features,
            "query_features": self.query_features,
        }

        for df_name, df in df_data.items():
            if df is not None:
                df_path = base_path / f"{df_name}.parquet"
                self._to_parquet(df, df_path)

    @classmethod
    def load(
        cls,
        path: str,
        dataframe_type: Optional[str] = None,
    ) -> Dataset:
        """
        Load the Dataset from the provided path.

        :param path: The file path
        :dataframe_type: Dataframe type to use to store internal data.
            Can be spark|pandas|polars|None.
            If not provided automatically sets to the one used when the Dataset was saved.
        :returns: Loaded Dataset.
        """
        base_path = Path(path).with_suffix(".replay").resolve()
        with open(base_path / "init_args.json") as file:
            dataset_dict = json.loads(file.read())

        if dataframe_type not in ["pandas", "spark", "polars", None]:
            msg = f"Argument dataframe_type can be spark|pandas|polars|None, not {dataframe_type}"
            raise ValueError(msg)

        feature_schema_data = dataset_dict["init_args"]["feature_schema"]
        features_list = []
        for feature_data in feature_schema_data:
            f_type = feature_data["feature_type"]
            f_hint = feature_data["feature_hint"]
            feature_data["feature_type"] = FeatureType[f_type] if f_type else None
            feature_data["feature_hint"] = FeatureHint[f_hint] if f_hint else None
            features_list.append(FeatureInfo(**feature_data))
        dataset_dict["init_args"]["feature_schema"] = FeatureSchema(features_list)

        for df_name in ["interactions", "query_features", "item_features"]:
            df_type = dataset_dict["init_args"][df_name]
            if df_type:
                df_type = dataframe_type or df_type
                load_path = base_path / f"{df_name}.parquet"
                dataset_dict["init_args"][df_name] = cls._read_parquet(load_path, df_type)
        dataset = cls(**dataset_dict["init_args"])
        return dataset

    if PYSPARK_AVAILABLE:

        def persist(self, storage_level: StorageLevel = StorageLevel(True, True, False, True, 1)) -> None:
            """
            Sets the storage level to persist SparkDataFrame for interactions, item_features
            and user_features.

            The function is only available when the PySpark is installed.

            :param storage_level: storage level to set for persistance.
                default: ```MEMORY_AND_DISK_DESER```.
            """
            if self.is_spark:
                self.interactions.persist(storage_level)
                if self.item_features is not None:
                    self.item_features.persist(storage_level)
                if self.query_features is not None:
                    self.query_features.persist(storage_level)

        def unpersist(self, blocking: bool = False) -> None:
            """
            Marks SparkDataFrame as non-persistent, and remove all blocks for it from memory and disk
            for interactions, item_features and user_features.

            The function is only available when the PySpark is installed.

            :param blocking: whether to block until all blocks are deleted.
                default: ```False```.
            """
            if self.is_spark:
                self.interactions.unpersist(blocking)
                if self.item_features is not None:
                    self.item_features.unpersist(blocking)
                if self.query_features is not None:
                    self.query_features.unpersist(blocking)

        def cache(self) -> None:
            """
            Persists the SparkDataFrame with the default storage level (MEMORY_AND_DISK)
            for interactions, item_features and user_features.

            The function is only available when the PySpark is installed.
            """
            if self.is_spark:
                self.interactions.cache()
                if self.item_features is not None:
                    self.item_features.cache()
                if self.query_features is not None:
                    self.query_features.cache()

    def subset(self, features_to_keep: Iterable[str]) -> Dataset:
        """
        Returns subset of features. Keeps query and item IDs even if
        the corresponding sources are not explicitly passed to this functions.

        :param features_to_keep: sequence of features to keep.

        :returns: new Dataset with given features.
        """
        # We always need to have query and item ID features in interactions dataset
        features_to_keep_set = set(features_to_keep)
        features_to_keep_set.add(self._feature_schema.query_id_column)
        features_to_keep_set.add(self._feature_schema.item_id_column)

        feature_schema_subset = self._feature_schema.subset(features_to_keep_set)

        interaction_fi = (
            FeatureSchema(feature_schema_subset.query_id_feature)
            + FeatureSchema(feature_schema_subset.item_id_feature)
            + feature_schema_subset.interaction_features
        )
        interactions = select(self._interactions, interaction_fi.columns)

        query_features = self._query_features
        if query_features is not None:
            query_fi = FeatureSchema(feature_schema_subset.query_id_feature) + feature_schema_subset.query_features
            query_features = select(query_features, query_fi.columns)

        item_features = self._item_features
        if item_features is not None:
            item_fi = FeatureSchema(feature_schema_subset.item_id_feature) + feature_schema_subset.item_features
            item_features = select(item_features, item_fi.columns)

        # We do not need to check consistency as it was already checked during parent dataset creation
        # Taking subset does not modify column values
        return Dataset(
            feature_schema=feature_schema_subset,
            interactions=interactions,
            query_features=query_features,
            item_features=item_features,
            check_consistency=False,
            categorical_encoded=self._categorical_encoded,
        )

    def _get_feature_source_map(self):
        self._feature_source_map: dict[FeatureSource, DataFrameLike] = {
            FeatureSource.INTERACTIONS: self.interactions,
            FeatureSource.QUERY_FEATURES: self.query_features,
            FeatureSource.ITEM_FEATURES: self.item_features,
        }

    def _get_ids_source_map(self):
        self._ids_feature_map: dict[FeatureHint, DataFrameLike] = {
            FeatureHint.QUERY_ID: self.query_features if self.query_features is not None else self.interactions,
            FeatureHint.ITEM_ID: self.item_features if self.item_features is not None else self.interactions,
        }

    def _assign_df_type(self):
        self.is_pandas = isinstance(self.interactions, PandasDataFrame)
        self.is_spark = isinstance(self.interactions, SparkDataFrame)
        self.is_polars = isinstance(self.interactions, PolarsDataFrame)

    def _get_cardinality(self, feature: FeatureInfo) -> Callable:
        def callback(column: str) -> int:
            if feature.feature_hint in [FeatureHint.ITEM_ID, FeatureHint.QUERY_ID]:
                return nunique(self._ids_feature_map[feature.feature_hint], column)
            assert feature.feature_source
            if feature.feature_type == FeatureType.CATEGORICAL_LIST:
                if self.is_spark:
                    data = (
                        self._feature_source_map[feature.feature_source]
                        .select(column)
                        .withColumn(column, sf.explode(column))
                    )
                else:
                    data = self._feature_source_map[feature.feature_source][[column]].explode(column)
                return nunique(data, column)
            return nunique(self._feature_source_map[feature.feature_source], column)

        return callback

    def _set_cardinality(self, features_list: Sequence[FeatureInfo]) -> None:
        for feature in features_list:
            if feature.feature_type in [FeatureType.CATEGORICAL, FeatureType.CATEGORICAL_LIST]:
                feature._set_cardinality_callback(self._get_cardinality(feature))

    def _fill_feature_schema(self, feature_schema: FeatureSchema) -> FeatureSchema:
        features_list = self._fill_unlabeled_features_sources(feature_schema=feature_schema)
        updated_feature_schema = FeatureSchema(features_list)

        filled_features = self._fill_unlabeled_features(
            source=FeatureSource.INTERACTIONS,
            feature_schema=updated_feature_schema,
        )

        if self.item_features is not None:
            filled_features += self._fill_unlabeled_features(
                source=FeatureSource.ITEM_FEATURES,
                feature_schema=updated_feature_schema,
            )

        if self.query_features is not None:
            filled_features += self._fill_unlabeled_features(
                source=FeatureSource.QUERY_FEATURES,
                feature_schema=updated_feature_schema,
            )
        return FeatureSchema(features_list=features_list + filled_features)

    def _fill_unlabeled_features_sources(self, feature_schema: FeatureSchema) -> list[FeatureInfo]:
        features_list = list(feature_schema.all_features)

        source_mapping: dict[str, FeatureSource] = {}
        for source in FeatureSource:
            dataframe = self._feature_source_map[source]
            if dataframe is not None:
                for column in dataframe.columns:
                    if column in feature_schema.columns:
                        source_mapping[column] = source

        for feature in features_list:
            if feature.feature_hint in [FeatureHint.QUERY_ID, FeatureHint.ITEM_ID]:
                feature._set_feature_source(source=FeatureSource.INTERACTIONS)
                continue
            source = source_mapping.get(feature.column)
            if source:
                feature._set_feature_source(source=source_mapping[feature.column])
            else:
                msg = f"{feature.column} doesn't exist in provided dataframes"
                raise ValueError(msg)

        self._set_cardinality(features_list=features_list)
        return features_list

    def _get_unlabeled_columns(self, source: FeatureSource, feature_schema: FeatureSchema) -> list[FeatureInfo]:
        set_source_dataframe_columns = set(self._feature_source_map[source].columns)
        set_labeled_dataframe_columns = set(feature_schema.columns)
        unlabeled_columns = set_source_dataframe_columns - set_labeled_dataframe_columns
        unlabeled_features_list = [
            FeatureInfo(column=column, feature_source=source, feature_type=FeatureType.NUMERICAL)
            for column in unlabeled_columns
        ]
        return unlabeled_features_list

    def _fill_unlabeled_features(self, source: FeatureSource, feature_schema: FeatureSchema) -> list[FeatureInfo]:
        unlabeled_columns = self._get_unlabeled_columns(source=source, feature_schema=feature_schema)
        self._set_features_source(feature_list=unlabeled_columns, source=source)
        self._set_cardinality(features_list=unlabeled_columns)
        return unlabeled_columns

    def _set_features_source(self, feature_list: list[FeatureInfo], source: FeatureSource) -> None:
        for feature in feature_list:
            feature._set_feature_source(source)

    def _check_ids_consistency(self, hint: FeatureHint) -> None:
        """
        Checks that all the ids from the interactions are in the features dataframe.
        """
        features_df = self._ids_feature_map[hint]
        ids_column = (
            self.feature_schema.item_id_column if hint == FeatureHint.ITEM_ID else self.feature_schema.query_id_column
        )
        if self.is_pandas:
            interactions_unique_ids = set(self.interactions[ids_column].unique())
            features_df_unique_ids = set(features_df[ids_column].unique())
            in_interactions_not_in_features_ids = interactions_unique_ids - features_df_unique_ids
            is_consistent = len(in_interactions_not_in_features_ids) == 0
        elif self.is_spark:
            is_consistent = (
                self.interactions.select(ids_column)
                .distinct()
                .join(
                    features_df.select(ids_column).distinct(),
                    on=[ids_column],
                    how="leftanti",
                )
                .count()
            ) == 0
        else:
            is_consistent = (
                len(
                    self.interactions.select(ids_column)
                    .unique()
                    .join(
                        features_df.select(ids_column).unique(),
                        on=ids_column,
                        how="anti",
                    )
                )
                == 0
            )

        if not is_consistent:
            msg = f"There are IDs in the interactions that are missing in the {hint.name} dataframe."
            raise ValueError(msg)

    def _check_column_encoded(
        self,
        data: DataFrameLike,
        column: str,
        source: FeatureSource,
        feature_type: FeatureType,
        cardinality: Optional[int],
    ) -> None:
        """
        Checks that IDs are encoded:
        1) IDs are integers;
        2) Min id >= 0;
        3) Max id < quantity of unique IDs.

        TODO: think about the third criterion. Case when full data was encoded and then splitted.
        Option: Keep this criterion, but suggest the user to disable the check if he understands
        that the criterion will not pass.
        """
        if feature_type == FeatureType.CATEGORICAL_LIST:  # explode column if list
            data = data.withColumn(column, sf.explode(column)) if self.is_spark else data[[column]].explode(column)

            if self.is_pandas:
                try:
                    data[column] = data[column].astype(int)
                except Exception as exc:
                    msg = f"IDs in {source.name}.{column} are not encoded. They are not int."
                    raise ValueError(msg) from exc

        if self.is_pandas:
            is_int = np.issubdtype(dict(data.dtypes)[column], int)
        elif self.is_spark:
            is_int = "int" in dict(data.dtypes)[column]
        else:
            is_int = data[column].dtype.is_integer()

        if not is_int:
            msg = f"IDs in {source.name}.{column} are not encoded. They are not int."
            raise ValueError(msg)

        if self.is_pandas:
            min_id = data[column].min()
        elif self.is_spark:
            min_id = data.agg(sf.min(column).alias("min_index")).first()[0]
        else:
            min_id = data[column].min()
        if min_id < 0:
            msg = f"IDs in {source.name}.{column} are not encoded. Min ID is less than 0."
            raise ValueError(msg)

        if self.is_pandas:
            max_id = data[column].max()
        elif self.is_spark:
            max_id = data.agg(sf.max(column).alias("max_index")).first()[0]
        else:
            max_id = data[column].max()

        if max_id >= cardinality:
            msg = f"IDs in {source.name}.{column} are not encoded. Max ID is more than quantity of IDs."
            raise ValueError(msg)

    def _check_encoded(self) -> None:
        for feature in self.feature_schema.categorical_features.all_features:
            if feature.feature_hint == FeatureHint.ITEM_ID:
                self._check_column_encoded(
                    self.interactions,
                    feature.column,
                    FeatureSource.INTERACTIONS,
                    feature.feature_type,
                    feature.cardinality,
                )
                if self.item_features is not None:
                    self._check_column_encoded(
                        self.item_features,
                        feature.column,
                        FeatureSource.ITEM_FEATURES,
                        feature.feature_type,
                        feature.cardinality,
                    )
            elif feature.feature_hint == FeatureHint.QUERY_ID:
                self._check_column_encoded(
                    self.interactions,
                    feature.column,
                    FeatureSource.INTERACTIONS,
                    feature.feature_type,
                    feature.cardinality,
                )
                if self.query_features is not None:
                    self._check_column_encoded(
                        self.query_features,
                        feature.column,
                        FeatureSource.QUERY_FEATURES,
                        feature.feature_type,
                        feature.cardinality,
                    )
            else:
                data = self._feature_source_map[feature.feature_source]
                self._check_column_encoded(
                    data,
                    feature.column,
                    feature.feature_source,
                    feature.feature_type,
                    feature.cardinality,
                )

    def to_pandas(self) -> None:
        """
        Convert internally stored dataframes to pandas.DataFrame.
        """
        from replay.utils.common import convert2pandas

        self._interactions = convert2pandas(self._interactions)
        if self._query_features is not None:
            self._query_features = convert2pandas(self._query_features)
        if self._item_features is not None:
            self._item_features = convert2pandas(self.item_features)
        self._get_feature_source_map()
        self._get_ids_source_map()
        self._assign_df_type()

    def to_spark(self):
        """
        Convert internally stored dataframes to pyspark.sql.DataFrame.
        """
        from replay.utils.common import convert2spark

        self._interactions = convert2spark(self._interactions)
        if self._query_features is not None:
            self._query_features = convert2spark(self._query_features)
        if self._item_features is not None:
            self._item_features = convert2spark(self._item_features)
        self._get_feature_source_map()
        self._get_ids_source_map()
        self._assign_df_type()

    def to_polars(self):
        """
        Convert internally stored dataframes to polars.DataFrame.
        """
        from replay.utils.common import convert2polars

        self._interactions = convert2polars(self._interactions)
        if self._query_features is not None:
            self._query_features = convert2polars(self._query_features)
        if self._item_features is not None:
            self._item_features = convert2polars(self._item_features)
        self._get_feature_source_map()
        self._get_ids_source_map()
        self._assign_df_type()


def nunique(data: DataFrameLike, column: str) -> int:
    """
    Returns number of unique values of specified column in dataframe.

    :param data: dataframe.
    :param column: column name.

    :returns: number of unique values.
    """
    if isinstance(data, SparkDataFrame):
        return data.select(column).distinct().count()
    if isinstance(data, PandasDataFrame):
        return data[column].nunique()
    return data.select(column).n_unique()


def select(data: DataFrameLike, columns: Sequence[str]) -> DataFrameLike:
    """
    :param data: dataframe.
    :param columns: sequence of column names to select.

    :returns: selected data in the same format as input dataframe.
    """
    if isinstance(data, SparkDataFrame):
        return data.select(*columns)
    if isinstance(data, PandasDataFrame):
        return data[columns]
    if isinstance(data, PolarsDataFrame):
        return data.select(*columns)
    assert False, "Unknown data frame type"


def check_dataframes_types_equal(dataframe: DataFrameLike, other: DataFrameLike):
    """
    :param dataframe: dataframe.
    :param other: dataframe to compare.

    :returns: True if dataframes have same type.
    """
    return any(
        [
            isinstance(dataframe, PandasDataFrame) and isinstance(other, PandasDataFrame),
            isinstance(dataframe, SparkDataFrame) and isinstance(other, SparkDataFrame),
            isinstance(dataframe, PolarsDataFrame) and isinstance(other, PolarsDataFrame),
        ]
    )



import logging
from collections.abc import Iterator
from typing import Optional

import numpy as np

from replay.models.extensions.ann.index_inferers.base_inferer import IndexInferer
from replay.models.extensions.ann.index_inferers.hnswlib_filter_index_inferer import HnswlibFilterIndexInferer
from replay.models.extensions.ann.index_inferers.hnswlib_index_inferer import HnswlibIndexInferer
from replay.models.extensions.ann.utils import create_hnswlib_index_instance
from replay.utils import PandasDataFrame, SparkDataFrame

from .base_index_builder import IndexBuilder

logger = logging.getLogger("replay")


class ExecutorHnswlibIndexBuilder(IndexBuilder):
    """
    Builder that build hnswlib index on one executor.
    """

    def produce_inferer(self, filter_seen_items: bool) -> IndexInferer:
        if filter_seen_items:
            return HnswlibFilterIndexInferer(self.index_params, self.index_store)
        else:
            return HnswlibIndexInferer(self.index_params, self.index_store)

    def build_index(
        self,
        vectors: SparkDataFrame,
        features_col: str,
        ids_col: Optional[str] = None,
    ):
        # to execution in one executor
        vectors = vectors.repartition(1)

        _index_store = self.index_store
        _index_params = self.index_params

        def build_index_udf(iterator: Iterator[PandasDataFrame]):  # pragma: no cover
            """Builds index on executor and writes it to shared disk or hdfs.

            Args:
                iterator: iterates on dataframes with vectors/features

            """
            index = create_hnswlib_index_instance(_index_params, init=True)

            # pdf is a pandas dataframe that contains ids and features (vectors)
            for pdf in iterator:
                vectors_np = np.squeeze(pdf[features_col].values)
                if ids_col:
                    index.add_items(np.stack(vectors_np), pdf[ids_col].values)
                else:
                    # ids will be from [0, ..., len(vectors_np)]
                    index.add_items(np.stack(vectors_np))

            _index_store.save_to_store(lambda path: index.save_index(path))

            yield PandasDataFrame(data={"_success": 1}, index=[0])

        # Here we perform materialization (`.collect()`) to build the hnsw index.
        cols = [ids_col, features_col] if ids_col else [features_col]

        vectors.select(*cols).mapInPandas(build_index_udf, "_success int").collect()
